from typing import Union, List, Tuple, Optional, Any, Dict, Set

import jsonpickle

from nexinfosys import ureg
from nexinfosys.command_generators import parser_field_parsers
from nexinfosys.command_generators.parser_ast_evaluators import ast_to_string
from nexinfosys.common.helper import PartialRetrievalDictionary, create_dictionary, strcmp, FloatOrString
from nexinfosys.model_services import State, get_case_study_registry_objects
from nexinfosys.models.musiasem_concepts import \
    FlowFundRoegenType, FactorInProcessorType, RelationClassType, \
    Processor, FactorType, Observer, Factor, \
    ProcessorsRelationPartOfObservation, ProcessorsRelationUndirectedFlowObservation, \
    ProcessorsRelationUpscaleObservation, \
    FactorsRelationDirectedFlowObservation, Hierarchy, Taxon, \
    FactorQuantitativeObservation, HierarchyLevel, Geolocation, FactorsRelationScaleObservation, \
    FactorTypesRelationUnidirectionalLinearTransformObservation
from nexinfosys.models.statistical_datasets import CodeList, CodeListLevel, Code


def serialize_hierarchy(h: Hierarchy) -> str:
    return jsonpickle.encode(h)


def deserialize_hierarchy(s: str) -> Hierarchy:
    return jsonpickle.decode(s)


def convert_code_list_to_hierarchy(cl: CodeList) -> Hierarchy:
    """
    It receives a CodeList and elaborates an equivalent Hierarchy, returning it

    :param cl: The input CodeList
    :return: The equivalent Hierarchy
    """
    h = Hierarchy(name=cl.code)
    h._description = cl.description
    # CodeList is organized in levels. Create all Levels and all Nodes (do not interlink Nodes)
    levels_dict = create_dictionary()
    code_node_dict = {}  # Maps a Code to the corresponding HierarchyNode
    for cll in cl.levels:
        hl = HierarchyLevel(cll.code, h)
        h.level_names.append(cll.code)
        h.levels.append(hl)
        levels_dict[cll.code] = hl
        for ct in cll.codes:
            hn = Taxon(ct.code, hierarchy=h, label=ct.description, description=ct.description)
            h.codes[ct.code] = hn
            code_node_dict[ct] = hn
            hn.level = levels_dict.get(ct.level.code, None)  # Point to the containing HierarchyLevel
            if hn.level:
                hn.level.codes.add(hn)

    # Set children & parents
    for ct in code_node_dict:
        p = code_node_dict[ct]
        for ch in ct.children:
            c = code_node_dict[ch]
            p._children.add(c)
            c._parents.append(p)
            c._parents_weights.append(1.0)

    # Finally, set "roots" to HierarchyNodes without "parents"
    tmp = []
    for c in h.codes.values():
        if len(c._parents) == 0:
            tmp.append(c)
    h.roots_append(tmp)

    return h


def convert_hierarchy_to_code_list(h: Hierarchy) -> CodeList:
    """
    It receives a Hierarchy and elaborates an equivalent CodeList, returning it

    NOTE: It does not support references to other Codes. Which means Codes member of multiple hierarchies.
          It does not support Codes with multiple parents and different weights.
          It does not support code lists with no levels. At least a level "" has to be defined (but this means it is a pure CodeList)

    :param h: The input Hierarchy
    :return: The equivalent CodeList
    """
    cl = CodeList()
    cl.code = h.name
    cl.description = h._description
    levels_map = {}
    # Levels
    for hl in h.levels:
        cll = CodeListLevel()
        cll.code_list = cl
        cll.code = hl.name
        levels_map[hl] = cll

    dummy_level = None

    # Codes
    codes_map = {}
    for hn in h.codes.values():
        c = Code()
        c.code = hn.name
        c.description = hn.description
        if hn.level:
            c.level = levels_map[hn.level]
        else:  # No level defined, assign a "dummy_level"
            if not dummy_level:
                dummy_level = CodeListLevel()
                dummy_level.code_list = cl
                dummy_level.code = ""
            c.level = dummy_level
        codes_map[hn] = c

    # Links between nodes
    for hn in h.codes.values():
        for ch in hn._children:
            p = codes_map[hn]
            c = codes_map[ch]
            p.children.append(c)  # Append a child
            c.parents.append(p)  # Add a parent

    return cl


def hierarchical_name_variants(h_name: str):
    """
    Given a hierarchical name, obtain variants

    :param h_name:
    :return: An ordered list of names that may refer to the same object
    """
    parts = h_name.split(".")
    if len(parts) > 1:
        return [h_name, parts[-1]]
    else:
        return [h_name]  # A simple name


def find_quantitative_observations(idx: PartialRetrievalDictionary, processor_instances_only=False) -> List[FactorQuantitativeObservation]:
    """ Find all available quantitative observations """
    interfaces: List[Factor] = idx.get(Factor.partial_key())

    return [o for i in interfaces
                    if not processor_instances_only or i.processor.instance_or_archetype == "Instance"
              for o in i.quantitative_observations
           ]


def find_observable_by_name(name: str, idx: PartialRetrievalDictionary, processor: Processor = None,
                            factor_type: FactorType = None) -> Union[Factor, Processor, FactorType]:
    """
    From a full Factor name "processor:factortype", obtain the corresponding Factor, searching in the INDEX of objects
    It supports also finding a Processor: "processor" (no ":factortype" part)
    It supports also finding a FactorType: ":factortype" (no "processor" part)
    It considers the fact that both Processors and FactorTypes can have different names
    (and consequently that Factors can have also multiple names)

    :param name: ":" separated processor name and factor type name. "p:ft" returns a Factor.
                 "p" or "p:" returns a Processor. ":ft" returns a FactorType
    :param idx: The PartialRetrievalDictionary where the objects have been previously indexed
    :param processor: Already resolved Processor. If ":ft" is specified, it will use this parameter to return a Factor
                 (not a FactorType)
    :param factor_type: Already resolved FactorType. If "p:" is specified (note the ":") , it will use this parameter
                 to return a Factor (not a Processor)
    :return: Processor or FactorType or Factor
    """
    res = None
    if isinstance(name, str):
        s = name.split(":")
        if len(s) == 2:  # There is a ":", so either FactorType or Factor (FactorType if there is no Processor)
            p_name = s[0]
            f_name = s[1]
            if not p_name:  # Processor can be blank
                p_name = None
            if not f_name:  # Factor type can be blank
                f_name = None
        elif len(s) == 1:  # If no ":", go just for the processor
            p_name = s[0]
            f_name = None
        # Retrieve the processor
        if p_name:
            for alt_name in hierarchical_name_variants(p_name):
                p = idx.get(Processor.partial_key(name=alt_name))
                if len(p) == 1:
                    p = p[0]
                    break
                elif len(p) > 1:
                    raise Exception(f"{alt_name} appears more than one time.")
        elif processor:
            p = processor
        else:
            p = None

        # Retrieve the FactorType
        if f_name:
            for alt_name in hierarchical_name_variants(f_name):
                ft = idx.get(FactorType.partial_key(name=alt_name))
                if ft:
                    ft = ft[0]
                    break
        elif factor_type:
            ft = factor_type
        else:
            res = p
            ft = None

        # Retrieve the Factor
        if not p_name and not p:  # If no Processor available at this point, FactorType is being requested, return it
            res = ft
        elif not res and p and ft:
            f = idx.get(Factor.partial_key(processor=p, factor_type=ft))
            if f:
                res = f[0]
    else:
        res = name

    return res


def find_processors_matching_name(processor_name, registry):
    parts, _ = obtain_name_parts(processor_name)
    ps = registry.get(Processor.partial_key(name=parts[0]))
    i = 1
    while len(ps) > 0 and i < len(parts):
        ps_tmp = []
        for p in ps:
            for partof_rel in registry.get(ProcessorsRelationPartOfObservation.partial_key(parent=p)):
                if strcmp(partof_rel.child_processor.name, parts[i]):
                    ps_tmp.append(partof_rel.child_processor)
                    break
        ps = ps_tmp
        i += 1
    return ps


def find_processor_by_name(state: Union[State, PartialRetrievalDictionary], processor_name: str) -> Optional[Processor]:
    """
    Find a processor by its name

    The name can be:
      * simple
      * partial hierarchical
      * absolute hierarchical

    The result can be:
      * None: no Processor,
      * a single Processor or
      * Exception if more than one Processor matches the name

    :param state:
    :param processor_name:
    :exception If more than one Processor can be found
    :return:
    """

    # Decompose the name
    p_names, _ = obtain_name_parts(processor_name)

    # Get registry object
    if isinstance(state, PartialRetrievalDictionary):
        glb_idx = state
    else:
        glb_idx, _, _, _, _ = get_case_study_registry_objects(state)

    if processor_name:
        ps = find_processors_matching_name(processor_name, glb_idx)
        if len(ps) == 1:
            return ps[0]  # One found!
        else:
            raise Exception(f"{len(ps)} processors matched '{processor_name}': {', '.join([p.full_hierarchy_names(glb_idx)[0] for p in ps])}")
    else:
        raise Exception("No processor name specified: '"+processor_name+"'")


def find_factortype_by_name(state: Union[State, PartialRetrievalDictionary], factortype_name: str):
    """
    Find FactorType by its name

    The name can be:
      * simple
      * absolute hierarchical

    The result can be:
      * None: no Factortype,
      * a single FactorType or
      * Exception if more than one FactorType matches the name

    :param state:
    :param factortype_name:
    :exception If more than one FactorType can be found
    :return:
    """

    # Decompose the name
    p_names, _ = obtain_name_parts(factortype_name)

    # Get registry object
    if isinstance(state, PartialRetrievalDictionary):
        glb_idx = state
    else:
        glb_idx, _, _, _, _ = get_case_study_registry_objects(state)

    if len(p_names) > 0:
        # Directly accessible
        ft = glb_idx.get(FactorType.partial_key(p_names[0]))
        if len(ft) == 1:
            ft = ft[0]
            if len(p_names) == 1:
                return ft[0]  # Found!
            else:
                # Look for children of "p" matching each piece
                for partial_name in p_names[1:]:
                    if len(ft.get_children()) == 0:
                        return None
                    else:
                        # Find child Processors matching the name
                        matches = []
                        for ftc in ft.get_children():
                            if strcmp(partial_name, ftc.name):
                                matches.append(ftc)
                        if len(matches) == 0:
                            return None
                        elif len(matches) == 1:
                            ft = matches[0]
                        else:
                            raise Exception(str(len(matches))+" InterfaceTypes matched '"+partial_name+"' in '"+factortype_name+"'")
        else:  # The number of matching top level FactorType is different from ONE
            if len(ft) == 0:
                return None
            else:
                raise Exception(str(len(ft)+" InterfaceTypes found matching '"+factortype_name+"'"))
    else:
        raise Exception("No InterfaceType name specified: '"+factortype_name+"'")


def find_or_create_observable(state: Union[State, PartialRetrievalDictionary],
                              name: str, source: Union[str, Observer]=Observer.no_observer_specified,
                              aliases: str=None,  # "name" (processor part) is an alias of "aliases" Processor
                              proc_attributes: Dict[str, Any] = None, proc_location: Geolocation = None,
                              fact_roegen_type: FlowFundRoegenType=None, fact_attributes: dict=None,
                              fact_incoming: bool=None, fact_external: bool=None, fact_location=None):
    """
    Find or create Observables: Processors, Factor and FactorType objects
    It can also create an Alias for a Processor if the name of the aliased Processor is passed (parameter "aliases")

    "name" is parsed, which can specify a processor AND a factor, both hierarchical ('.'), separated by ":"

    :param state:
    :param name: Full name of processor, processor':'factor or ':'factor
    :param source: Name of the observer or Observer itself (used only when creating nested Processors, because it implies part-of relations)
    :param aliases: Full name of an existing processor to be aliased by the processor part in "name"
    :param proc_attributes: Dictionary with attributes to be added to the processor if it is created
    :param proc_location: Specification of where the processor is physically located, if it applies
    :param fact_roegen_type: Flow or Fund
    :param fact_attributes: Dictionary with attributes to be added to the Factor if it is created
    :param fact_incoming: True if the Factor is incoming regarding the processor; False if it is outgoing
    :param fact_external: True if the Factor comes from Environment
    :param fact_location: Specification of where the processor is physically located, if it applies
    :return: Processor, FactorType, Factor
    """

    # Decompose the name
    p_names, f_names = obtain_name_parts(name)

    # Get objects from state
    if isinstance(state, PartialRetrievalDictionary):
        glb_idx = state
    else:
        glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)

    # Get the Observer for the relations (PART-OF for now)
    if source:
        if isinstance(source, Observer):
            oer = source
        else:
            oer = glb_idx.get(Observer.partial_key(name=source))
            if not oer:
                oer = Observer(source)
                glb_idx.put(oer.key(), oer)
            else:
                oer = oer[0]

    p = None  # Processor to which the Factor is connected
    ft = None  # FactorType
    f = None  # Factor

    if p_names and aliases:
        # Create an alias for the Processor
        if isinstance(aliases, str):
            p = glb_idx.get(Processor.partial_key(aliases))
        elif isinstance(aliases, Processor):
            p = aliases
        if p:
            full_name = ".".join(p_names)
            # Look for a processor named <full_name>, it will be an AMBIGUITY TO BE AVOIDED
            p1, k1 = glb_idx.get(Processor.partial_key(full_name), True)
            if p1:
                # If it is an ALIAS, void the already existing because there would be no way to decide
                # which of the two (or more) do we need
                if Processor.is_alias_key(k1[0]):
                    # Assign NONE to the existing Alias
                    glb_idx.put(k1[0], None)
            else:
                # Create the ALIAS
                k_ = Processor.alias_key(full_name, p)
                glb_idx.put(k_, p)  # An alternative Key pointing to the same processor
    else:
        # Find or create the "lineage" of Processors, using part-of relation ("matryoshka" or recursive containment)
        parent = None
        acum_name = ""
        for i, p_name in enumerate(p_names):
            last = i == (len(p_names)-1)

            # CREATE processor(s) (if it does not exist). The processor is an Observable
            acum_name += ("." if acum_name != "" else "") + p_name
            p = glb_idx.get(Processor.partial_key(name=acum_name))
            if not p or strcmp(acum_name, p_name):
                attrs = proc_attributes if last else None
                location = proc_location if last else None
                # Create processor
                p = Processor(p_name,  # acum_name,
                              geolocation=location,
                              tags=None,
                              attributes=attrs
                              )
                # Index it, with its multiple names, adding the attributes only if it is the processor in play
                # for alt_name in hierarchical_name_variants(acum_name):
                p_key = Processor.partial_key(acum_name, p.ident)
                if last and proc_attributes:
                    p_key.update({k: ("" if v is None else v) for k, v in proc_attributes.items()})
                glb_idx.put(p_key, p)
            else:
                p = p[0]

            if parent:
                # Create PART-OF relation
                o1 = glb_idx.get(ProcessorsRelationPartOfObservation.partial_key(parent=parent, child=p))
                if not o1:
                    o1 = ProcessorsRelationPartOfObservation.create_and_append(parent, p, oer)  # Part-of
                    glb_idx.put(o1.key(), o1)
                    p_key = Processor.partial_key(f"{parent.name}.{acum_name}", p.ident)
                    if proc_attributes:
                        p_key.update({k: ("" if v is None else v) for k, v in proc_attributes.items()})
                    glb_idx.put(p_key, p)

            parent = p

    # Find or create the lineage of FactorTypes and for the last FactorType, find or create Factor
    parent = None
    acum_name = ""
    for i, ft_name in enumerate(f_names):
        last = i == len(f_names)-1

        # CREATE factor type(s) (if it does not exist). The Factor Type is a Class of Observables
        # (it is NOT observable: neither quantities nor relations)
        acum_name += ("." if acum_name != "" else "") + ft_name
        ft = glb_idx.get(FactorType.partial_key(name=acum_name))
        if not ft:
            attrs = fact_attributes if last else None
            ft = FactorType(acum_name,  #
                            parent=parent, hierarchy=None,
                            roegen_type=fact_roegen_type,
                            tags=None,  # No tags
                            attributes=attrs,
                            expression=None  # No expression
                            )
            for alt_name in hierarchical_name_variants(acum_name):
                ft_key = FactorType.partial_key(alt_name, ft.ident)
                if last and fact_attributes:
                    ft_key.update(fact_attributes)
                glb_idx.put(ft_key, ft)
        else:
            ft = ft[0]

        if last and p:  # The Processor must exist. If not, nothing is created or obtained
            # CREATE Factor (if it does not exist). An Observable
            f = glb_idx.get(Factor.partial_key(processor=p, factor_type=ft))
            if not f:
                f = Factor.create_and_append(acum_name,
                                             p,
                                             in_processor_type=FactorInProcessorType(external=fact_external, incoming=fact_incoming),
                                             taxon=ft,
                                             geolocation=fact_location,
                                             tags=None,
                                             attributes=fact_attributes)
                glb_idx.put(f.key(), f)
            else:
                f = f[0]

        parent = ft

    return p, ft, f  # Return all the observables (some may be None)


def find_or_create_factor(state: Union[State, PartialRetrievalDictionary],
                          p: Processor, ft: FactorType,
                          fact_external: bool=None, fact_incoming: bool=None,
                          fact_location=None, fact_attributes: dict=None):
    if isinstance(state, PartialRetrievalDictionary):
        glb_idx = state
    else:
        glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)

    f = glb_idx.get(Factor.partial_key(processor=p, factor_type=ft))
    if not f:
        f = Factor.create_and_append(ft.name,
                                     p,
                                     in_processor_type=FactorInProcessorType(external=fact_external,
                                                                             incoming=fact_incoming),
                                     taxon=ft,
                                     geolocation=fact_location,
                                     tags=None,
                                     attributes=fact_attributes)
        glb_idx.put(f.key(), f)
    else:
        f = f[0]
    return f


def find_or_create_factor_type(state: Union[State, PartialRetrievalDictionary],
                               name: str,
                               fact_roegen_type: FlowFundRoegenType = None, fact_attributes: dict = None):
    """
    "name" has to be a FactorType name, ie, ":<hierarchical_name>"

    :param state:
    :param name:
    :param fact_roegen_type:
    :param fact_attributes:
    :return:
    """
    p_names, f_names = obtain_name_parts(name)
    if not p_names:
        _, ft, _ = find_or_create_observable(state, name, fact_roegen_type=fact_roegen_type,
                                             fact_attributes=fact_attributes)
        return ft
    else:
        raise Exception("It has to be a FactorType name, received '" + name + "'")


def find_or_create_processor(state: Union[State, PartialRetrievalDictionary],
                             name: str,
                             proc_attributes: dict = None,
                             proc_location=None):
    """
    "name" has to be a Processor name

    :param state:
    :param name:
    :param proc_attributes:
    :param proc_location:
    :return:
    """
    p_names, f_names = obtain_name_parts(name)
    if not f_names:
        p, _, _ = find_or_create_observable(state, name,
                                            proc_attributes=proc_attributes,
                                            proc_location=proc_location)
        return p
    else:
        raise Exception("It has to be a Processor name, received '" + name + "'")


def create_or_append_quantitative_observation(state: Union[State, PartialRetrievalDictionary],
                                              factor: Union[str, Factor],
                                              value: str, unit: str,
                                              observer: Union[str, Observer]=Observer.no_observer_specified,
                                              spread: str=None, assessment: str=None, pedigree: str=None, pedigree_template: str=None,
                                              relative_to: Union[str, Factor]=None,
                                              time: str=None,
                                              geolocation: str=None,
                                              comments: str=None,
                                              tags=None, other_attributes=None,
                                              proc_aliases: str=None,
                                              proc_external: bool=None, proc_attributes: dict=None, proc_location=None,
                                              ftype_roegen_type: FlowFundRoegenType=None, ftype_attributes: dict=None,
                                              fact_incoming: bool=None, fact_external: bool=None, fact_location=None):
    """
    Creates an Observation of a Factor
    If the containing Processor does not exist, it is created
    If the FactorType does not exist, it is created
    If the Factor does not exist, it is created
    Finally, if no "value" is passed, only the Factor is created

    :param state: A State or PartialRetrievalDictionary instance
    :param factor: string processor:factor_type or Factor. If str, the Factor (and Processor and FactorType) can be created
    :param value: expression with the value
    :param unit: metric unit
    :param observer: string with the name of the observer or Observer
    :param spread: expression defining uncertainty of :param value
    :param assessment:
    :param pedigree: encoded assessment of the quality of the science/technique of the observation
    :param pedigree_template: reference pedigree matrix used to encode the pedigree
    :param relative_to: Factor Type in the same Processor to which the value is relative
    :param time: time extent in which the value is valid
    :param geolocation: where the observation is
    :param comments: open comments about the observation
    :param tags: list of tags added to the observation
    :param other_attributes: dictionary added to the observation
    :param proc_aliases: name of aliased processor (optional). Used only if the Processor does not exist
    :param proc_external: True if the processor is outside the case study borders, False if it is inside. Used only if the Processor does not exist
    :param proc_attributes: Dictionary with attributes added to the Processor. Used only if the Processor does not exist
    :param proc_location: Reference specifying the location of the Processor. Used only if the Processor does not exist
    :param ftype_roegen_type: Either FUND or FLOW (applied to FactorType). Used only if the FactorType does not exist
    :param ftype_attributes: Dictionary with attributes added to the FactorType. Used only if the FactorType does not exist
    :param fact_incoming: Specifies if the Factor goes into or out the Processor. Used if the Factor (not FactorType) does not exist
    :param fact_external: Specifies if the Factor is injected from an external Processor. Used if the Factor (not FactorType) does not exist
    :param fact_location: Reference specifying the location of the Factor. Used if the Factor does not exist

    :return:
    """
    # Get objects from state
    if isinstance(state, State):
        glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)
    elif isinstance(state, PartialRetrievalDictionary):
        glb_idx = state

    # Obtain factor
    p, ft = None, None
    if not isinstance(factor, Factor):
        p, ft, factor_ = find_or_create_observable(glb_idx,
                                                   factor,
                                                   # source=None,
                                                   aliases=proc_aliases,
                                                   proc_attributes=proc_attributes,
                                                   proc_location=proc_location,
                                                   fact_roegen_type=ftype_roegen_type,
                                                   fact_attributes=ftype_attributes,
                                                   fact_incoming=fact_incoming,
                                                   fact_external=fact_external,
                                                   fact_location=fact_location
                                                   )
        if not isinstance(factor_, Factor):
            raise Exception("The name specified for the factor ('"+factor+"') did not result in the obtention of a Factor")
        else:
            factor = factor_

    # If a value is defined...
    if value:
        # Get the Observer for the relations (PART-OF for now)
        if isinstance(observer, Observer):
            oer = observer
        else:
            if not observer:
                observer = Observer.no_observer_specified
            oer = glb_idx.get(Observer.partial_key(name=observer))
            if not oer:
                oer = Observer(observer)
                glb_idx.put(oer.key(), oer)
            else:
                oer = oer[0]

        # If "relative_to" is specified, maybe a FactorType needs to be created
        if relative_to:
            ast = parser_field_parsers.string_to_ast(parser_field_parsers.factor_unit, relative_to)
            factor_type = ast_to_string(ast["factor"])

            unit_name = ast["unparsed_unit"]
            unit = str((ureg(unit) / ureg(unit_name)).units)

            ft = find_or_create_factor_type(glb_idx, ":"+factor_type, fact_roegen_type=None, fact_attributes=None)
            relative_to = find_or_create_factor(glb_idx, p, ft, fact_external=None, fact_incoming=None,
                                                fact_location=None, fact_attributes=None)

        # Create the observation (and append it to the Interface)
        o = _create_or_append_quantitative_observation(factor,
                                                       value, unit, spread, assessment, pedigree, pedigree_template,
                                                       oer,
                                                       relative_to,
                                                       time,
                                                       geolocation,
                                                       comments,
                                                       tags, other_attributes
                                                       )
        # Register
        # glb_idx.put(o.key(), o)

        # Return the observation
        return p, ft, factor, o
    else:
        # Return the Factor
        return p, ft, factor, None


def create_relation_observations(state: Union[State, PartialRetrievalDictionary],
                                 origin: Union[str, Processor, Factor],
                                 destinations: List[Tuple[Union[str, Processor, Factor], Optional[Tuple[Union[RelationClassType, str], Optional[str]]]]],
                                 relation_class: Union[str, RelationClassType]=None,
                                 oer: Union[str, Observer]=Observer.no_observer_specified,
                                 attributes=None) -> List:
    """
    Create and register one or more relations from a single origin to one or more destinations.
    Relation parameters (type and weight) can be specified for each destination, or a default relation class parameter is used
    Relation are assigned to the observer "oer"

    :param state: Registry of all objects
    :param origin: Origin of the relation as string, Processor or Factor
    :param destinations: List of tuples, where each tuple can be of a single element, the string, Processor or Factor, or can be accompanied by the relation parameters, the relation type, and the string specifying the weight
    :param relation_class: Default relation class
    :param oer: str or Observer for the Observer to which relation observations are accounted
    :param attributes: Attributes attached to the new Relationship
    :return: The list of relations
    """
    def get_requested_object(p_, ft_, f_):
        return f_ if f_ else (p_ if p_ else ft_)

    def get_all_objects(input: Union[str, Processor, Factor]) -> Tuple[Processor, Optional[FactorType], Optional[Factor]]:
        if isinstance(input, str):
            return find_or_create_observable(glb_idx, input)
        elif isinstance(input, Processor):
            return input, None, None
        elif isinstance(input, Factor):
            return input.processor, input.taxon, input
        else:
            raise Exception(f"Input parameter must be String, Processor or Interface: type={type(input)}, value={input}")

    if isinstance(state, PartialRetrievalDictionary):
        glb_idx = state
    else:
        glb_idx, _, _, _, _ = get_case_study_registry_objects(state)

    # Origin
    p, ft, f = get_all_objects(origin)

    origin_obj = get_requested_object(p, ft, f)

    rels = []

    # Default Observer
    if not oer:
        oer = Observer.no_observer_specified

    # Create Observer
    if isinstance(oer, str):
        oer_ = glb_idx.get(Observer.partial_key(name=oer))
        if not oer_:
            oer = Observer(oer)
            glb_idx.put(oer.key(), oer)
        else:
            oer = oer_[0]
    elif not isinstance(oer, Observer):
        raise Exception("'oer' parameter must be a string or an Observer instance")

    if not isinstance(destinations, list):
        destinations = [destinations]

    for dst in destinations:
        if not isinstance(dst, tuple):
            dst = tuple([dst])
        # Destination
        dst_obj = None
        # PART-OF
        if isinstance(origin_obj, Processor) and relation_class == RelationClassType.pp_part_of:
            # Find dst[0]. If it does not exist, create dest UNDER (hierarchically) origin
            dst_obj = find_observable_by_name(dst[0], glb_idx)
            if not dst_obj:
                name = origin_obj.full_hierarchy_names(glb_idx)[0] + "." + dst[0]
                p, ft, f = find_or_create_observable(glb_idx, name, source=oer)
                dst_obj = get_requested_object(p, ft, f)
            rel = glb_idx.get(ProcessorsRelationPartOfObservation.partial_key(parent=origin_obj, child=dst_obj))  # , observer=oer
            if len(rel) == 0:
                rel = _find_or_create_relation(origin_obj, dst_obj, rel_type, oer, "", glb_idx, attributes=attributes)
            rels.append(rel[0])
            continue  # Skip the rest of the loop

        # not PART-OF relationship types
        if not dst_obj:
            p, ft, f = get_all_objects(dst[0])

            dst_obj = get_requested_object(p, ft, f)
            # If origin is Processor and destination is Factor, create Factor in origin (if it does not exist). Or viceversa
            if isinstance(origin_obj, Processor) and isinstance(dst_obj, Factor):
                # Obtain full origin processor name
                names = origin_obj.full_hierarchy_names(glb_idx)
                p, ft, f = find_or_create_observable(glb_idx, names[0] + ":" + dst_obj.taxon.name)
                origin_obj = get_requested_object(p, ft, f)
            elif isinstance(origin_obj, Factor) and isinstance(dst_obj, Processor):
                names = dst_obj.full_hierarchy_names(glb_idx)
                p, ft, f = find_or_create_observable(glb_idx, names[0] + ":" + origin_obj.taxon.name)
                dst_obj = get_requested_object(p, ft, f)
            # Relation class
            if len(dst) > 1:
                rel_type = dst[1]
            else:
                if not relation_class:
                    if isinstance(origin_obj, Processor) and isinstance(dst_obj, Processor):
                        relation_class = RelationClassType.pp_undirected_flow
                    else:
                        relation_class = RelationClassType.ff_directed_flow
                rel_type = relation_class
            if len(dst) > 2:
                weight = dst[2]
            else:
                weight = ""  # No weight, it only can be used to aggregate
            rel = _find_or_create_relation(origin_obj, dst_obj, rel_type, oer, weight, glb_idx, attributes=attributes)
        rels.append(rel)

    return rels

# ########################################################################################
# Auxiliary functions
# ########################################################################################


def obtain_name_parts(n):
    """
    Parse the name. List of processor names + list of factor names
    :param n:
    :return:
    """
    r = n.split(":")
    if len(r) > 1:
        full_p_name = r[0]
        full_f_name = r[1]
    else:
        full_p_name = r[0]
        full_f_name = ""
    p_ = full_p_name.split(".")
    f_ = full_f_name.split(".")
    if len(p_) == 1 and not p_[0]:
        p_ = []
    if len(f_) == 1 and not f_[0]:
        f_ = []
    return p_, f_


def get_factor_id(f_: Union[Factor, Processor], ft: FactorType=None):
    if isinstance(f_, Factor):
        return (f_.processor.name + ":" + f_.taxon.name).lower()
    elif isinstance(f_, Processor) and isinstance(ft, FactorType):
        return (f_.name + ":" + ft.name).lower()


def _create_or_append_quantitative_observation(factor: Factor,
                                               value: str, unit: str,
                                               spread: str, assessment: str, pedigree: str, pedigree_template: str,
                                               observer: Observer,
                                               relative_to: Optional[Factor],
                                               time: str,
                                               geolocation: Optional[str],
                                               comments: str,
                                               tags, other_attributes):
    f_name = get_factor_id(factor)
    # print(f_name)
    if other_attributes:
        attrs = other_attributes.copy()
    else:
        attrs = {}

    attrs.update({"relative_to": relative_to,
                  "time": time,
                  "geolocation": geolocation,
                  "unit": unit,
                  "spread": spread,
                  "assessment": assessment,
                  "pedigree": pedigree,
                  "pedigree_template": pedigree_template,
                  "comments": comments
                  }
                 )

    fo = FactorQuantitativeObservation.create_and_append(v=value,
                                                         factor=factor,
                                                         observer=observer,
                                                         tags=tags,
                                                         attributes=attrs
                                                         )
    return fo


def _get_observer(observer: Union[str, Observer], registry: PartialRetrievalDictionary) -> Optional[Observer]:
    res = None
    if isinstance(observer, Observer):
        res = observer
    else:
        oer = registry.get(Observer.partial_key(name=observer))
        if oer:
            res = oer[0]
    return res


def find_or_create_observer(observer: str, registry: PartialRetrievalDictionary) -> Observer:
    # Find
    obs = _get_observer(observer, registry)

    if not obs:
        # Create
        obs = Observer(observer)
        registry.put(obs.key(), obs)

    return obs


def find_factor_types_transform_relation(registry: PartialRetrievalDictionary,
                                         origin_interface_type: FactorType, destination_interface_type: FactorType,
                                         origin_context: Processor, destination_context: Processor):
    """We try to get the best match from the existing factor types scale changes"""

    def get_relations_from_contexts(orig=None, dest=None) -> List[FactorTypesRelationUnidirectionalLinearTransformObservation]:
        return registry.get(FactorTypesRelationUnidirectionalLinearTransformObservation.partial_key(
            origin=origin_interface_type, destination=destination_interface_type,
            origin_context=orig, destination_context=dest))

    # 1. Do we have a relation with same origin and destination context?
    if origin_context and destination_context:
        relations = get_relations_from_contexts(orig=origin_context, dest=destination_context)
        if len(relations) > 0:
            return relations

    # 2. Do we have a relation with same origin context?
    if origin_context:
        relations = get_relations_from_contexts(orig=origin_context)
        if len(relations) > 0:
            return relations

    # 3. Do we have a relation with same destination context?
    if destination_context:
        relations = get_relations_from_contexts(dest=destination_context)
        if len(relations) > 0:
            return relations

    # 4. Do we have a relation without contexts?
    return get_relations_from_contexts()


def _find_or_create_relation(origin, destination, rel_type: Union[str, RelationClassType], oer: Union[Observer, str], weight: str, state: Union[State, PartialRetrievalDictionary], attributes=None):
    """
    Construct and register a relation between origin and destination

    :param origin: Either processor or factor
    :param destination: Either processor or factor
    :param rel_type: Relation type. Either a string or a member of RelationClassType enumeration
    :param oer: Observer, as object or string
    :param weight: For flow relations
    :param state: State or PartialRetrievalDictionary
    :param attributes: Attributes of the relationship
    :return: The relation observation
    """
    # Get objects from state
    if isinstance(state, State):
        glb_idx, _, _, _, _ = get_case_study_registry_objects(state)
    elif isinstance(state, PartialRetrievalDictionary):
        glb_idx = state

    # CREATE the Observer for the relation
    if oer and isinstance(oer, str):
        oer_ = glb_idx.get(Observer.partial_key(name=oer))
        if not oer_:
            oer = Observer(oer)
            glb_idx.put(oer.key(), oer)
        else:
            oer = oer_[0]

    if isinstance(rel_type, str):
        rel_type = RelationClassType.from_str(rel_type)

    r = None
    if rel_type.is_between_processors and isinstance(origin, Processor) and isinstance(destination, Processor):
        if rel_type == RelationClassType.pp_part_of:
            # Find or Create the relation
            r = glb_idx.get(ProcessorsRelationPartOfObservation.partial_key(parent=origin, child=destination))
            if not r:
                r = ProcessorsRelationPartOfObservation.create_and_append(origin, destination, oer, attributes=attributes)  # Part-of
                glb_idx.put(r.key(), r)
            else:
                r = r[0]
            # Add destination to the index with an alternative name
            # TODO Do the same with all part-of children of destination, recursively
            # TODO "full_hierarchy_names" makes use of
            d_name = destination.simple_name()
            for h_name in origin.full_hierarchy_names(glb_idx):
                full_name = h_name+"."+d_name
                p = glb_idx.get(Processor.partial_key(name=full_name))
                if not p:
                    glb_idx.put(Processor.partial_key(name=full_name, ident=destination.ident), destination)
                else:
                    if p[0].ident != destination.ident:
                        raise Exception("Two Processors under name '"+full_name+"' have been found: ID1: "+p[0].ident+"; ID2: "+destination.ident)
        elif rel_type == RelationClassType.pp_undirected_flow:
            # Find or Create the relation
            r = glb_idx.get(ProcessorsRelationUndirectedFlowObservation.partial_key(source=origin, target=destination))
            if not r:
                r = ProcessorsRelationUndirectedFlowObservation.create_and_append(origin, destination, oer, attributes=attributes)  # Undirected flow
                glb_idx.put(r.key(), r)
            else:
                r = r[0]
        elif rel_type == RelationClassType.pp_upscale:
            # Find or Create the relation
            r = glb_idx.get(ProcessorsRelationUpscaleObservation.partial_key(parent=origin, child=destination))
            if not r:
                r = ProcessorsRelationUpscaleObservation.create_and_append(origin, destination, oer, weight, attributes=attributes)  # Upscale
                glb_idx.put(r.key(), r)
            else:
                r = r[0]
                r._quantity = weight
                r._observer = oer
                r._attributes = attributes
    elif rel_type.is_between_interfaces and isinstance(origin, Factor) and isinstance(destination, Factor):

        if rel_type in (RelationClassType.ff_directed_flow, RelationClassType.ff_reverse_directed_flow,
                        RelationClassType.ff_directed_flow_back):
            if rel_type == RelationClassType.ff_reverse_directed_flow:
                origin, destination = destination, origin

                if weight:
                    weight = f"1/({weight})"

                scale_change_weight = attributes.get("scale_change_weight") if attributes else None
                if scale_change_weight:
                    attributes["scale_change_weight"] = f"1/({scale_change_weight})"

            back_interface: Optional[Factor] = attributes.pop("back_interface", None) if attributes else None

            # Find or Create the relation
            r = glb_idx.get(FactorsRelationDirectedFlowObservation.partial_key(source=origin,
                                                                               target=destination,
                                                                               back=back_interface))
            if not r:
                r = FactorsRelationDirectedFlowObservation.create_and_append(origin, destination, oer, weight,
                                                                             attributes=attributes,
                                                                             back=back_interface)
                glb_idx.put(r.key(), r)
            else:
                r = r[0]
                r._weight = weight
                r._observer = oer
                r._attributes = attributes

        elif rel_type in (RelationClassType.ff_scale, RelationClassType.ff_scale_change):
            scale_change_weight = attributes.pop("scale_change_weight", None) if attributes else None
            if scale_change_weight is not None:
                weight = FloatOrString.multiply(weight, scale_change_weight)

            # Find or Create the relation
            r = glb_idx.get(FactorsRelationScaleObservation.partial_key(origin=origin, destination=destination))
            if not r:
                r = FactorsRelationScaleObservation.create_and_append(origin, destination, oer, weight,
                                                                      attributes=attributes)
                glb_idx.put(r.key(), r)
            else:
                r = r[0]
                r._quantity = weight
                r._observer = oer
                r._attributes = attributes

    return r


def build_hierarchy(name, type_name, registry: PartialRetrievalDictionary, h: dict, oer: Observer=None, level_names=None):
    """
    Take the result of parsing a hierarchy and elaborate either an Hierarchy (for Categories and FactorType)
    or a set of nested Processors

    Shortcut function

    :param name:
    :param type_name:
    :param registry:
    :param h:
    :param oer: An Observer
    :param level_names:
    :return: If type_name is Processor -> "None". If other, return the Hierarchy object
    """
    if type_name.lower() in ["p"]:
        type_name = "processor"
    elif type_name.lower() in ["i", "f"]:
        type_name = "factortype"
    elif type_name.lower() in ["c", "t"]:
        type_name = "taxon"

    return _build_hierarchy(name, type_name, registry, h, oer, level_names, acum_name="", parent=None)


def _build_hierarchy(name, type_name, registry: PartialRetrievalDictionary, h: dict, oer=None, level_names=None, acum_name="", parent=None) -> Optional[Hierarchy]:
    """
    Take the result of parsing a hierarchy and elaborate either an Hierarchy (for Categories and FactorType)
    or a set of nested Processors

    :param name: Name of the hierarchy
    :param type_name: Processor, Taxonomy, FactorType
    :param registry: The state, space of variables where the nodes and the hierarchy itself are stored
    :param h: The list of nodes, which can be recursive
    :param oer: The observer of the hierarchy. It is treated differently for Processors, Categories and FactorTypes
    :param level_names:
    :param acum_name: (Internal - do not use) Will contain the acumulated name in hierarchical form
    :param parent: Parent to be used to define the relations in the current level of the hierarchy
    :return:
    """
    # Get or create hierarchy or observer
    if oer:
        hie = oer
    elif name:
        if type_name.lower() == "processor":
            if registry:
                oer = registry.get(Observer.partial_key(name=name))
                if oer:
                    hie = oer[0]
                else:
                    hie = Observer(name)
                    registry.put(hie.key(), hie)
            else:
                hie = Observer(name)
        else:
            if registry:
                hie = registry.get(Hierarchy.partial_key(name=name))
                if not hie:
                    hie = Hierarchy(name, type_name=type_name)
                    registry.put(hie.key(), hie)
                else:
                    hie = hie[0]
            else:
                hie = Hierarchy(name, type_name=type_name)

    for s in h:
        # Create node
        n_name = s["code"]
        if "description" in s:
            desc = s["description"]
        else:
            desc = None
        if "expression" in s:
            exp = s["expression"]
        else:
            exp = None
        if "children" in s:
            children = s["children"]
        else:
            children = []
        # Accumulated name
        acum_name2 = acum_name + ("." if acum_name != "" else "") + n_name

        if type_name.lower() == "processor":
            # Check if the Processor exists
            # If not, create it
            n = registry.get(Processor.partial_key(name=acum_name2))
            if not n:
                attrs = None
                location = None
                proc_external = None
                n = Processor(acum_name2,
                              geolocation=location,
                              tags=None,
                              attributes=attrs
                              )
                registry.put(n.key(), n)
            else:
                n = n[0]
            if parent:
                # Create "part-of" relation
                rel = _find_or_create_relation(parent, n, RelationClassType.pp_part_of, hie, "", registry)
        elif type_name.lower() == "factortype":
            # Check if the FactorType exists
            # If not, create it
            n = registry.get(FactorType.partial_key(name=acum_name2))
            if not n:
                attrs = None
                n = FactorType(acum_name2,  #
                               parent=parent, hierarchy=hie,
                               roegen_type=FlowFundRoegenType.flow,
                               tags=None,  # No tags
                               attributes=attrs,
                               expression=exp
                               )
                for alt_name in hierarchical_name_variants(acum_name2):
                    ft_key = FactorType.partial_key(alt_name, n.ident)
                    registry.put(ft_key, n)
            else:
                n = n[0]
        elif type_name.lower() == "taxon":
            # Check if the Taxon exists
            # If not, create it
            n = registry.get(Taxon.partial_key(name=acum_name2))
            if not n:
                n = Taxon(acum_name2, parent=parent, hierarchy=hie, expression=exp, description=desc)
                for alt_name in hierarchical_name_variants(acum_name2):
                    t_key = Taxon.partial_key(alt_name, n.ident)
                    registry.put(t_key, n)
            else:
                n = n[0]

        if children:
            _build_hierarchy(name, type_name, registry, children, hie, level_names, acum_name2, parent=n)

        # Add node, only for the first level, if it is a hierarchy
        if not parent and isinstance(hie, Hierarchy):
            hie.roots_append(n)

    # Set level names and return the hierarchy, only for the first level, if it is a hierarchy
    if not parent and isinstance(hie, Hierarchy):
        if level_names:
            hie.level_names = level_names

        return hie
    else:
        return None


def get_processors(registry: PartialRetrievalDictionary):
    # Just remove duplicates (processors can be registered multiple times under different names)
    return set(registry.get(Processor.partial_key()))



from nexinfosys.command_generators import parser_field_parsers
from nexinfosys.common.helper import create_dictionary


def parse_line(item, fields):
    """
    Convert fields from a line to AST

    :param item:
    :param fields:
    :return:
    """
    asts = {}
    for f, v in item.items():
        if not f.startswith("_"):
            field = fields[f]
            # Parse (success is guaranteed because of the first pass dedicated to parsing)
            asts[f] = parser_field_parsers.string_to_ast(field.parser, v)
    return asts


def classify_variables(asts, datasets, hierarchies, parameters):
    """
    Iterate through the variables in a AST, determining
    for each if it is a:
    * dataset
    * hierarchy
    * parameter
    * or unknown

    :param asts: The AST
    :param datasets: The collection of Datasets
    :param hierarchies: The collection of Category Hierarchies
    :param parameters: The collection of Parameters
    :return: A dictionary with the classification,
    """
    ds = set()
    ds_concepts = set()
    hh = set()
    params = set()
    not_classif = set()
    for ast in asts:
        for var in ast["variables"] if "variables" in ast else []:
            parts = var.split(".")
            first_part = parts[0]
            if first_part in datasets:
                ds.add(datasets[first_part])
                ds_concepts.add(parts[1])  # Complete concept name: dataset"."concept
            elif first_part in hierarchies:
                hh.add(hierarchies[first_part])
            elif first_part in parameters:
                params.add(first_part)
            else:
                not_classif.add(first_part)

    return dict(datasets=ds, ds_concepts=ds_concepts, hierarchies=hh, parameters=params, not_classif=not_classif)


def classify_variables2(h_names, datasets, hierarchies, parameters):
    """
    Iterate through the hierarchical names in "h_names", determining for each if it is a:
        * dataset,
        * hierarchy,
        * parameter, or
        * unknown

    :param h_names: List of hierarchical names
    :param datasets: The collection of Datasets
    :param hierarchies: The collection of Category Hierarchies
    :param parameters: The collection of Parameters
    :return: A dictionary with the classification,
    """
    ds = set()
    ds_concepts = set()
    hh = set()
    params = set()
    not_classif = set()
    for var in h_names:
        parts = var.split(".")
        first_part = parts[0]
        if first_part in datasets:
            ds.add(datasets[first_part])
            ds_concepts.add(parts[1])  # Complete concept name: dataset"."concept
        elif first_part in hierarchies:
            hh.add(hierarchies[first_part])
        elif first_part in parameters:
            params.add(first_part)
        else:
            not_classif.add(first_part)

    return dict(datasets=list(ds), ds_concepts=list(ds_concepts), hierarchies=list(hh), parameters=list(params), not_classif=list(not_classif))


def obtain_dictionary_with_literal_fields(item, asts):
    d = create_dictionary()
    for f in item:
        if not f.startswith("_"):
            ast = asts[f]
            if "complex" not in ast or ("complex" in ast and not ast["complex"]):
                d[f] = item[f]
    return d



import geojson
import urllib

from nexinfosys.common.helper import create_dictionary
from nexinfosys.model_services import get_case_study_registry_objects
from nexinfosys.models.musiasem_concepts import Processor, GeographicReference

in_files = create_dictionary()  # URL -> (json, idx)


def read_geojson(url):
    """
    Read a GeoJSON file and index it by ID

    :param url:
    :return: A tuple with the deserialized GeoJSON file and an index of ID to position in the features list
    """
    if url not in in_files:
        f = urllib.request.urlopen(url)
        j = geojson.loads(f.read())
        id_dict = create_dictionary()
        for i, f in enumerate(j["features"]):
            fid = f["id"]
            id_dict[fid] = i
        in_files[url] = (j, id_dict)
    else:
        j, id_dict = in_files[url]

    return j, id_dict


def generate_geojson(s):
    """

    :param s:
    :return:
    """

    features = []
    # Obtain principal structures
    glb_idx, p_sets, hierarchies, datasets, mappings = get_case_study_registry_objects(s)
    ps = glb_idx.get(Processor.partial_key())
    for p in ps:
        # TODO If there is Geolocation information, obtain the region
        #      Collect also the attributes of the processor and of all the interfaces, from "flow graph solution"
        # url = "https://raw.githubusercontent.com/eurostat/Nuts2json/master/2016/4326/20M/nutsrg_1.json"
        if not p.geolocation:
            continue

        gr = glb_idx.get(GeographicReference.partial_key(name=p.geolocation.reference))
        url = gr[0].attributes["data_location"]
        geo_id = p.geolocation.code
        j, ids = read_geojson(url)
        # Obtain element of interest (GeoJSON)
        tmp = j["features"][ids[geo_id]]
        feature = tmp.copy()
        # Add processor properties
        feature["properties"]["processor"] = dict(name=p.name, h_name=p.full_hierarchy_names(glb_idx)[0], subsystem_type=p.subsystem_type, system=p.processor_system, level=p.level, functional_or_structural=p.functional_or_structural, instance_or_archetype=p.instance_or_archetype, stock=p.stock)
        # Add interface properties
        interfaces = []
        for i in p.factors:
            idict = dict()
            interfaces.append(idict)
        feature["properties"]["interfaces"] = interfaces
        # TODO Add relationships?

        features.append(feature)

    return dict(type="FeatureCollection", features=features)





__version__ = '11.7.17'
# See license text in pyparsing.py

# Modifications by Glenn Linderman, same license
#
# Python 3.2
# Eliminate LE and friends, stick with Python comparison ops
# Merge separate precedence lists
# Make a single class Arith to be the interface
# Keep vars_ in Arith instances, so multiple instances could have
#   different vars/values
# Add // and % to multOp & EvalMultOp
# Keep integer values as integers until something converts them
# Allow longer var names

# Based on:
#
# eval_arith.py
#
# Copyright 2009, Paul McGuire
#
# Expansion on the pyparsing example simpleArith.py, to include evaluation
# of the parsed tokens.

from pyparsing import Word, nums, alphas, Combine, oneOf, Optional, \
    opAssoc, operatorPrecedence


class EvalConstant():
    "Class to evaluate a parsed constant or variable"

    def __init__(self, tokens):
        self.value = tokens[0]

    def eval(self, vars_):
        if self.value in vars_:
            return vars_[self.value]
        else:
            try:
                return int(self.value)
            except:
                return float(self.value)


class EvalSignOp():
    "Class to evaluate expressions with a leading + or - sign"

    def __init__(self, tokens):
        self.sign, self.value = tokens[0]

    def eval(self, vars_):
        mult = {'+': 1, '-': -1}[self.sign]
        return mult * self.value.eval(vars_)


def operatorOperands(tokenlist):
    "generator to extract operators and operands in pairs"
    it = iter(tokenlist)
    while 1:
        try:
            o1 = next(it)
            o2 = next(it)
            yield (o1, o2)
        except StopIteration:
            break


class EvalMultOp():
    "Class to evaluate multiplication and division expressions"

    def __init__(self, tokens):
        self.value = tokens[0]

    def eval(self, vars_):
        prod = self.value[0].eval(vars_)
        for op, val in operatorOperands(self.value[1:]):
            if op == '*':
                prod *= val.eval(vars_)
            if op == '/':
                prod /= val.eval(vars_)
            if op == '//':
                prod //= val.eval(vars_)
            if op == '%':
                prod %= val.eval(vars_)
        return prod


class EvalAddOp():
    "Class to evaluate addition and subtraction expressions"

    def __init__(self, tokens):
        self.value = tokens[0]

    def eval(self, vars_):
        sum = self.value[0].eval(vars_)
        for op, val in operatorOperands(self.value[1:]):
            if op == '+':
                sum += val.eval(vars_)
            if op == '-':
                sum -= val.eval(vars_)
        return sum


class EvalComparisonOp():
    "Class to evaluate comparison expressions"
    opMap = {
        "<": lambda a, b: a < b,
        "<=": lambda a, b: a <= b,
        ">": lambda a, b: a > b,
        ">=": lambda a, b: a >= b,
        "==": lambda a, b: a == b,
        "!=": lambda a, b: a != b,
    }

    def __init__(self, tokens):
        self.value = tokens[0]

    def eval(self, vars_):
        val1 = self.value[0].eval(vars_)
        for op, val in operatorOperands(self.value[1:]):
            fn = self.opMap[op]
            val2 = val.eval(vars_)
            if not fn(val1, val2):
                break
            val1 = val2
        else:
            return True
        return False


class Arith():
    # define the parser
    integer = Word(nums)
    real = (Combine(Word(nums) + Optional("." + Word(nums))
                    + oneOf("E e") + Optional(oneOf('+ -')) + Word(nums))
            | Combine(Word(nums) + "." + Word(nums))
            )

    variable = Word(alphas)
    operand = real | integer | variable

    signop = oneOf('+ -')
    multop = oneOf('* / // %')
    plusop = oneOf('+ -')
    comparisonop = oneOf("< <= > >= == != <>")

    # use parse actions to attach EvalXXX constructors to sub-expressions
    operand.setParseAction(EvalConstant)
    arith_expr = operatorPrecedence(operand,
                                    [(signop, 1, opAssoc.RIGHT, EvalSignOp),
                                     (multop, 2, opAssoc.LEFT, EvalMultOp),
                                     (plusop, 2, opAssoc.LEFT, EvalAddOp),
                                     (comparisonop, 2, opAssoc.LEFT, EvalComparisonOp),
                                     ])

    def __init__(self, vars_={}):
        self.vars_ = vars_

    def setvars(self, vars_):
        self.vars_ = vars_

    def setvar(var, val):
        self.vars_[var] = val

    def eval(self, strExpr):
        ret = self.arith_expr.parseString(strExpr, parseAll=True)[0]
        result = ret.eval(self.vars_)
        return result


if __name__ == '__main__':
    greet = Word(alphas) + "," + Word(alphas) + "!"

    # input string
    hello = "Hello, World!"

    # parse input string
    print(hello + "->" + str(greet.parseString(hello)))

    def main():
        # sample expressions posted on comp.lang.python, asking for advice
        # in safely evaluating them
        rules = [
            '( A - B ) == 0',
            '(A + B + C + D + E + F + G + H + I) == J',
            '(A + B + C + D + E + F + G + H) == I',
            '(A + B + C + D + E + F) == G',
            '(A + B + C + D + E) == (F + G + H + I + J)',
            '(A + B + C + D + E) == (F + G + H + I)',
            '(A + B + C + D + E) == F',
            '(A + B + C + D) == (E + F + G + H)',
            '(A + B + C) == (D + E + F)',
            '(A + B) == (C + D + E + F)',
            '(A + B) == (C + D)',
            '(A + B) == (C - D + E - F - G + H + I + J)',
            '(A + B) == C',
            '(A + B) == 0',
            '(A+B+C+D+E) == (F+G+H+I+J)',
            '(A+B+C+D) == (E+F+G+H)',
            '(A+B+C+D)==(E+F+G+H)',
            '(A+B+C)==(D+E+F)',
            '(A+B)==(C+D)',
            '(A+B)==C',
            '(A-B)==C',
            '(A/(B+C))',
            '(B/(C+D))',
            '(G + H) == I',
            '-0.99 <= ((A+B+C)-(D+E+F+G)) <= 0.99',
            '-0.99 <= (A-(B+C)) <= 0.99',
            '-1000.00 <= A <= 0.00',
            '-5000.00 <= A <= 0.00',
            'A < B',
            'A < 7000',
            'A == -(B)',
            'A == C',
            'A == 0',
            'A > 0',
            'A > 0.00',
            'A > 7.00',
            'A <= B',
            'A < -1000.00',
            'A < -5000',
            'A < 0',
            'A==(B+C+D)',
            'A==B',
            'I == (G + H)',
            '0.00 <= A <= 4.00',
            '4.00 < A <= 7.00',
            '0.00 <= A <= 4.00 <= E > D',
            '123E0 > 1000E-1 > 99.0987',
            '123E+0',
            '1000E-1',
            '99.0987',
            'abc',
            '20 % 3',
            '14 // 3',
            '12e2 // 3.7',
        ]
        vars_ = {'A': 0, 'B': 1.1, 'C': 2.2, 'D': 3.3, 'E': 4.4, 'F': 5.5, 'G':
            6.6, 'H': 7.7, 'I': 8.8, 'J': 9.9, "abc": 20, }

        # define tests from given rules
        tests = []
        for t in rules:
            tests.append((t, eval(t, vars_)))

        # copy vars_ to EvalConstant lookup dict
        arith = Arith(vars_)
        for test, expected in tests:
            result = arith.eval(test)
            print(test, expected, result)
            # print( ret )
            if expected != result:
                print("*" * 60)


    main()



import json
from nexinfosys.model_services import IExecutableCommand


class PedigreeMatricesReferencesCommand(IExecutableCommand):
    def __init__(self, name: str):
        self._name = name
        self._content = None

    def execute(self, state: "State"):
        some_error = False
        issues = []

        return None, None

    def estimate_execution_time(self):
        return 0

    def json_serialize(self):
        # Directly return the metadata dictionary
        return self._content

    def json_deserialize(self, json_input):
        # TODO Check validity
        issues = []
        if isinstance(json_input, dict):
            self._content = json_input
        else:
            self._content = json.loads(json_input)

        if "description" in json_input:
            self._description = json_input["description"]
        return issues


import io
import logging

import networkx as nx

from nexinfosys.solving import *


def construct_processors_graph_2(state: State, query: IQueryObjects, filt: Union[str, dict], part_of: bool, scale: bool, flow: bool, format: str):
    """

    :param state:
    :param query:
    :param filt:
    :param part_of: show part_of relationships
    :param scale: show scale relationships (between processors)
    :param flow: show flow relationships
    :return:
    """

    def append_processor(proc, nodes, p_dict):
        p_id = get_processor_ident(proc)
        d = processor_to_dict(proc, reg)
        d.update(dict(graphics={}))
        # Instance / archetype
        if proc.instance_or_archetype.lower() == "instance":
            color = "#aa2211"
        else:
            color = "#552211"
        # Biosphere / technosphere
        if proc.subsystem_type.lower() in ["environment", "localenvironment", "externalenvironment"]:
            shape = "ellipse"
        else:
            shape = "hexagon"
        # Internal / external
        if proc.subsystem_type.lower() in ["external", "externalenvironment"]:  # Internal / external
            d["shape_properties"] = dict(borderDashes="[5,5]")
        d["color"] = color
        d["type"] = shape
        p_dict[p_id] = d
        # print(F"{p_id} -> {d}")
        nodes.append((p_id, d))

    # Format
    edge_flow = dict(graphics={"fill": "#ff00ff", "width": 3, "targetArrow": "standard"})
    edge_part_of = dict(graphics={"fill": "#00ffff", "width": 3, "targetArrow": "standard"})
    edge_scale = dict(graphics={"fill": "#00ffff", "width": 2, "targetArrow": "standard"})

    # Obtain the information needed to elaborate the graph
    objs = query.execute([Processor, Factor,
                          ProcessorsRelationPartOfObservation,
                          FactorsRelationScaleObservation,
                          FactorsRelationDirectedFlowObservation
                          ],
                         filt
                         )

    # PartialRetrievalDictionary
    reg = state.get("_glb_idx")

    processors = {}
    n = []
    e = []
    # Processors
    for p in objs[Processor]:
        # nam = " ,".join(p.full_hierarchy_names(reg))
        # print(nam)
        append_processor(p, n, processors)

    if flow:
        # Flow Links
        for flow in objs[FactorsRelationDirectedFlowObservation]:
            sf = get_factor_type_id(flow.source_factor)
            tf = get_factor_type_id(flow.target_factor)
            weight = flow.weight if flow.weight else ""
            weight = ("\n" + str(weight)) if weight != "" else ""
            sp_id = get_processor_ident(flow.source_factor.processor)
            dp_id = get_processor_ident(flow.target_factor.processor)
            if sf == tf:
                edge = dict(etype="flow", font={"size": 7}, label=sf + weight, w=weight)
            else:
                edge = dict(etype="flow", font={"size": 7}, label=sf + weight + "\n" + tf, w=weight)
            edge.update(edge_flow)
            e.append((sp_id, dp_id, edge))

    if part_of:
        # Part-of Relations
        for po in objs[ProcessorsRelationPartOfObservation]:
            # Add edge between two Processors
            pp = get_processor_ident(po.parent_processor)
            cp = get_processor_ident(po.child_processor)
            if pp not in processors:
                append_processor(po.parent_processor, n, processors)
            if cp not in processors:
                append_processor(po.child_processor, n, processors)
            edge = dict(etype="part-of", font={"size": 7})
            edge.update(edge_part_of)
            # print(F"{cp} -> {pp} [{edge}]")
            e.append((cp, pp, edge))

    if scale:
        # Scale Relations
        for scale in objs[FactorsRelationScaleObservation]:
            # Add edge between two Processors
            pp = get_processor_ident(scale.origin.processor)
            cp = get_processor_ident(scale.destination.processor)
            if pp not in processors:
                print("Not registered invoker")
            if cp not in processors:
                print("Not registered invoked")
            weight = scale.quantity if scale.quantity else ""
            weight = str(weight) if weight != "" else ""
            edge = dict(etype="scale", font={"size": 7}, dashes="true", label=scale.origin.name + " x " + weight +" > " + scale.destination.name, w=weight)
            edge.update(edge_scale)
            # print(F"{pp} -> {cp} [{edge}]")
            e.append((pp, cp, edge))

    # NetworkX
    # -- Create the graph
    processors_graph = nx.MultiDiGraph()
    processors_graph.add_nodes_from(n)
    processors_graph.add_edges_from(e)

    # Convert to VisJS
    ids_map = create_dictionary()
    id_count = 0
    for node in processors_graph.nodes(data=True):
        sid = str(id_count)
        node[1]["id"] = sid
        ids_map[node[0]] = sid
        id_count += 1

    ret = None
    if format == "visjs":
        vis_nodes = []
        vis_edges = []
        for node in processors_graph.nodes(data=True):
            d2 = node[1]
            logging.debug(node)
            if "uname" not in d2:
                continue
            d = dict(id=node[1]["id"], label=node[1]["uname"])
            if "shape" in node[1]:
                # circle, ellipse, database, box, diamond, dot, square, triangle, triangleDown, text, star
                d["shape"] = node[1]["shape"]
            else:
                d["shape"] = "hexagon"
            if "color" in node[1]:
                d["color"] = node[1]["color"]
            vis_nodes.append(d)
        for edge in processors_graph.edges(data=True):
            f = ids_map[edge[0]]
            t = ids_map[edge[1]]
            d = {"from": f, "to": t, "arrows": "to"}
            data = edge[2]
            if "label" in data:
                d["font"] = {"align": "middle"}
            d.update(data)

            vis_edges.append(d)
        ret = {"nodes": vis_nodes, "edges": vis_edges}
    elif format == "gml":
        ret1 = io.BytesIO()
        nx.write_gml(processors_graph, ret1)
        ret = ret1.getvalue()
        ret1.close()

    return ret

def construct_processors_graph(state: State, query: IQueryObjects, filt: Union[str, dict]):
    """
    Prepare a graph representing processors and their interconnections
    Optionally represent the relation "part-of" between Processors

    :param state: State
    :param query: A IQueryObjects instance (which has been already injected the state)
    :param filt: A filter to be passed to the query instance
    :return:
    """
    # Format
    a_processor = dict(graphics={"type": "hexagon", "color": "#aa2211"})
    edge_flow = dict(relation="flow", graphics={"fill": "#ff00ff", "width": 3, "targetArrow": "standard"})
    edge_part_of = dict(relation="part_of", graphics={"fill": "#00ffff", "width": 3, "targetArrow": "standard"})

    # Obtain the information needed to elaborate the graph
    objs = query.execute([Processor, Factor,
                          ProcessorsRelationPartOfObservation,
                          FactorsRelationDirectedFlowObservation
                          ],
                         filt
                         )

    # PartialRetrievalDictionary
    reg = state.get("_glb_idx")

    n = []
    e = []
    # Processors
    for p in objs[Processor]:
        # nam = " ,".join(p.full_hierarchy_names(reg))
        # print(nam)

        p_id = get_processor_ident(p)
        n.append((p_id, processor_to_dict(p, reg)))

    # Flow Links
    for flow in objs[FactorsRelationDirectedFlowObservation]:
        sf = get_factor_type_id(flow.source_factor)
        tf = get_factor_type_id(flow.target_factor)
        weight = flow.weight if flow.weight else ""
        weight = ("\n" + str(weight)) if weight != "" else ""
        sp_id = get_processor_ident(flow.source_factor.processor)
        dp_id = get_processor_ident(flow.target_factor.processor)
        if sf == tf:
            edge = dict(font={"size": 7}, label=sf + weight, w=str(weight))
        else:
            edge = dict(font={"size": 7}, label=sf + weight + "\n" + tf, w=str(weight))
        edge.update(edge_flow)
        e.append((sp_id, dp_id, edge))

    # Part-of Relations
    for po in objs[ProcessorsRelationPartOfObservation]:
        # TODO add edge between two Processors
        pp = get_processor_ident(po.parent_processor)
        cp = get_processor_ident(po.child_processor)
        edge = dict(font={"size": 7})
        edge.update(edge_part_of)
        e.append((cp, pp, edge))

    # NetworkX
    # -- Create the graph
    processors_graph = nx.MultiDiGraph()
    processors_graph.add_nodes_from(n)
    processors_graph.add_edges_from(e)

    # Convert to VisJS
    ids_map = create_dictionary()
    id_count = 0
    for node in processors_graph.nodes(data=True):
        sid = str(id_count)
        node[1]["id"] = sid
        ids_map[node[0]] = sid
        id_count += 1

    vis_nodes = []
    vis_edges = []
    for node in processors_graph.nodes(data=True):
        d = dict(id=node[1]["id"], label=node[1]["uname"])
        if "shape" in node[1]:
            # circle, ellipse, database, box, diamond, dot, square, triangle, triangleDown, text, star
            d["shape"] = node[1]["shape"]
        else:
            d["shape"] = "hexagon"
        if "color" in node[1]:
            d["color"] = node[1]["color"]
        vis_nodes.append(d)
    for edge in processors_graph.edges(data=True):
        f = ids_map[edge[0]]
        t = ids_map[edge[1]]
        d = {"from": f, "to": t, "arrows": "to"}
        data = edge[2]
        if "label" in data:
            d["label"] = data["label"]
            d["font"] = {"align": "middle"}
            if "font" in data:
                d["font"].update(data["font"])

        vis_edges.append(d)
    visjs = {"nodes": vis_nodes, "edges": vis_edges}
    # print(visjs)
    return visjs


if __name__ == '__main__':
    from nexinfosys.serialization import deserialize_state

    # Deserialize previously recorded Soslaires State (WARNING! Execute unit tests to generated the ".serialized" file)
    fname = "/home/rnebot/GoogleDrive/AA_MAGIC/Soslaires.serialized"
    fname = "/home/rnebot/GoogleDrive/AA_MAGIC/MiniAlmeria.serialized"
    with open(fname, "r") as file:
        s = file.read()
    state = deserialize_state(s)
    # Create a Query and execute a query
    query = BasicQuery(state)
    logging.debug(construct_processors_graph(state, query, None))



import json
import logging
from collections import OrderedDict
import traceback
import numpy as np
import pandas as pd
from typing import List

import nexinfosys
from nexinfosys.common.helper import strcmp, create_dictionary, obtain_dataset_metadata, \
    translate_case

from nexinfosys.models import CodeImmutable
from nexinfosys.models.statistical_datasets import Dataset, Dimension, CodeList

from nexinfosys.model_services import IExecutableCommand, get_case_study_registry_objects


# TODO Currently is just a copy of "ETLExternalDatasetCommand"
# TODO The new command has to elaborate a MDataset
# TODO It has two new parameters: "InputDataset" and "AvailableAtDateTime"
# TODO Time dimension can be specified as "Time" or as "StartTime" "EndTime"
# TODO Result parameter column also change a bit
# TODO See "DatasetQry" command in "MuSIASEM case study commands" Google Spreadsheet
#
from nexinfosys.models.musiasem_concepts_helper import convert_code_list_to_hierarchy


def obtain_reverse_codes(mapped, dst):
    """
    Given the list of desired dst codes and an extensive map src -> dst,
    obtain the list of src codes

    :param mapped: Correspondence between src codes and dst codes [{"o", "to": [{"d", "e"}]}]
    :param dst: Iterable of destination codes
    :return: List of origin codes
    """
    src = set()
    dest_set = set([d.lower() for d in dst])  # Destination categories
    # Obtain origin categories referencing "dest_set" destination categories
    for k in mapped:
        for t in k["to"]:
            if t["d"] and t["d"].lower() in dest_set:
                src.add(k["o"])
    return list(src)  # list(set([k[0].lower() for k in mapped if k[1].lower() in dest_set]))


def pctna(x):
    """
    Aggregation function computing the percentage of NaN values VS total number of elements, in a group "x"
    """
    return 100.0 * np.count_nonzero(np.isnan(x.astype(np.float))) / x.size


class DatasetQryCommand(IExecutableCommand):
    def __init__(self, name: str):
        self._name = name
        self._content = None

    def _create_new_dataset(self, name: str, in_ds: Dataset, df: pd.DataFrame, dimension_names: List[str], measure_names: List[str]):
        """
        Create a dataset using the original ds
        :param in_ds:
        :param df:
        :return:
        """
        ds = Dataset()

        ds.data = df
        ds.code = name
        ds.description = "Result of DatasetQry command"
        ds.attributes = {}
        ds.metadata = None
        ds.database = None

        for dimension in dimension_names:  # type: str
            d = Dimension()
            d.code = dimension
            d.description = None
            d.attributes = None
            d.is_time = "period" in dimension.lower()
            d.is_measure = False
            cl = df[dimension].unique().tolist()
            d.code_list = CodeList.construct(
                dimension, dimension, [""],
                codes=[CodeImmutable(c, c, "", []) for c in cl]
            )
            d.dataset = ds

        for measure in measure_names:  # type: str
            d = Dimension()
            d.code = measure
            d.description = None
            d.attributes = None
            d.is_time = False
            d.is_measure = True
            d.code_list = None
            d.dataset = ds

        return ds

    def execute(self, state: "State"):
        """
        First bring the data considering the filter
        Second, group, third aggregate
        Finally, store the result in State
        """
        issues = []
        # Obtain global variables in state
        glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)

        # DS Source + DS Name
        source = self._content["dataset_source"]
        dataset_name = self._content["dataset_name"]
        dataset_datetime = self._content["dataset_datetime"]

        # Result name
        result_name = self._content["result_name"]
        if result_name in datasets or state.get(result_name):
            issues.append((2, "A dataset called '"+result_name+"' is already stored in the registry of datasets"))

        # Dataset metadata
        dims, attrs, measures = obtain_dataset_metadata(dataset_name, source, datasets)

        # Obtain filter parameters
        params = create_dictionary()  # Native dimension name to list of values the filter will allow to pass
        joined_dimensions = []
        for dim in self._content["where"]:
            lst = self._content["where"][dim]
            native_dim = None
            if dim.lower() in ["startperiod", "starttime", "endperiod", "endtime"]:
                native_dim = dim
                lst = [lst]
            elif dim not in dims:
                # Check if there is a mapping. If so, obtain the native equivalent(s). If not, ERROR
                for m in mappings:
                    if strcmp(mappings[m].destination, dim) and \
                            strcmp(mappings[m].source, source) and \
                            strcmp(mappings[m].dataset, dataset_name) and \
                            mappings[m].origin in dims:
                        joined_dimensions.append(mappings[m].destination)  # Store dimension in the original case
                        native_dim = mappings[m].origin
                        lst = obtain_reverse_codes(mappings[m].map, lst)
                        break
            else:
                # Get the dimension name with the original case
                native_dim = dims[dim].name
            if native_dim:
                if native_dim not in params:
                    f = set()
                    params[native_dim] = f
                else:
                    f = params[native_dim]
                f.update(lst)

        # Convert param contents from set to list
        for p in params:
            params[p] = [i for i in params[p]]

        # Obtain the filtered Dataset <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
        ds = nexinfosys.data_source_manager.get_dataset_filtered(source, dataset_name, params, datasets)
        df = ds.data

        # Join with mapped dimensions (augment it)
        mapping_dict = create_dictionary()
        for m in mappings:
            if strcmp(mappings[m].source, source) and \
                    strcmp(mappings[m].dataset, dataset_name) and \
                    mappings[m].origin in dims:
                # mapping_tuples.append((mappings[m].origin, mappings[m].destination, mappings[m].map))
                mapping_dict[mappings[m].origin] = (mappings[m].destination, {d["o"]: d["to"] for d in mappings[m].map})

        # If accelerated version not available, use slow version
        try:
            if nexinfosys.get_global_configuration_variable("ENABLE_CYTHON_OPTIMIZATIONS") == "True":
                from nexinfosys.restful_service.helper_accel import augment_dataframe_with_mapped_columns2 as augment_df
            else:
                raise Exception("Just to import the slow version")
        except:
            from nexinfosys.common.helper import augment_dataframe_with_mapped_columns as augment_df

        df = augment_df(df, mapping_dict, ["value"])

        # Aggregate (If any dimension has been specified)
        if len(self._content["group_by"]) > 0:
            # Column names where data is
            # HACK: for the case where the measure has been named "obs_value", use "value"
            values = [m.lower() if m.lower() != "obs_value" else "value" for m in self._content["measures"]]
            v2 = []
            for v in values:
                for c in df.columns:
                    if v.lower() == c.lower():
                        v2.append(c)
                        break
            values = v2

            # TODO: use metadata name (e.g. "OBS_VALUE") instead of hardcoded "value"
            # values = self._content["measures"]
            out_names = self._content["measures_as"]
            group_by_dims = translate_case(self._content["group_by"], df.columns)  # Group by dimension names
            lcase_group_by_dims = [d.lower() for d in group_by_dims]
            # Now joined_dimensions
            for d in joined_dimensions:
                if d.lower() in lcase_group_by_dims:
                    # Find and replace
                    for i, d2 in enumerate(group_by_dims):
                        if strcmp(d, d2):
                            group_by_dims[i] = d
                            break

            agg_funcs = []  # Aggregation functions
            agg_names = {}
            for f in self._content["agg_funcs"]:
                if f.lower() in ["avg", "average"]:
                    agg_funcs.append(np.average)
                    agg_names[np.average] = "avg"
                elif f.lower() in ["sum"]:
                    agg_funcs.append(np.sum)
                    agg_names[np.sum] = "sum"
                elif f.lower() in ["count"]:
                    agg_funcs.append(np.size)
                    agg_names[np.size] = "count"
                elif f.lower() in ["sumna"]:
                    agg_funcs.append(np.nansum)
                    agg_names[np.nansum] = "sumna"
                elif f.lower() in ["countav"]:
                    agg_funcs.append("count")
                    agg_names["count"] = "countav"
                elif f.lower() in ["avgna"]:
                    agg_funcs.append(np.nanmean)
                    agg_names[np.nanmean] = "avgna"
                elif f.lower() in ["pctna"]:
                    agg_funcs.append(pctna)
                    agg_names[pctna] = "pctna"

            # Calculate Pivot Table. The columns are a combination of values x aggregation functions
            # For instance, if two values ["v2", "v2"] and two agg. functions ["avg", "sum"] are provided
            # The columns will be: [["average", "v2"], ["average", "v2"], ["sum", "v2"], ["sum", "v2"]]
            try:
                # Check that all "group_by_dims" on which pivot table aggregates are present in the input "df"
                # If not either synthesize them (only if there is a single filter value) or remove (if not present
                for r in group_by_dims.copy():
                    df_columns_dict = create_dictionary(data={c: None for c in df.columns})
                    if r not in df_columns_dict:
                        found = False
                        for k in params:
                            if strcmp(k, r):
                                found = True
                                if len(params[k]) == 1:
                                    df[k] = params[k][0]
                                else:
                                    group_by_dims.remove(r)
                                    issues.append((2, "Dimension '" + r + "' removed from the list of dimensions because it is not present in the raw input dataset."))
                                break
                        if not found:
                            group_by_dims.remove(r)
                            issues.append((2, "Dimension '" + r + "' removed from the list of dimensions because it is not present in the raw input dataset."))

                # Create and register Hierarchy objects from origin Dataset dimensions: state, ds
                ds_columns_dict = create_dictionary(data={c.code: c.code for c in ds.dimensions})
                for r in group_by_dims:
                    if r in ds_columns_dict:
                        # Create hierarchy local to the dataset
                        for d in ds.dimensions:
                            if strcmp(r, d.code):
                                if d.code_list:
                                    h = convert_code_list_to_hierarchy(d.code_list)
                                    h.name = result_name + "_" + r
                                    glb_idx.put(h.key(), h)
                                    break

                # Pivot table using Group by
                if True:
                    groups = df.groupby(by=group_by_dims, as_index=False)  # Split
                    d = OrderedDict([])
                    lst_names = []
                    if len(values) == len(agg_funcs):
                        for i, (value, agg_func) in enumerate(zip(values, agg_funcs)):
                            if len(out_names) == len(values) and out_names[i]:
                                lst_names.append(out_names[i])
                            else:
                                lst_names.append(agg_names[agg_func] + "_" +value)
                            lst = d.get(value, [])
                            lst.append(agg_func)
                            d[value] = lst
                    else:
                        for value in values:
                            lst = d.get(value, [])
                            for agg_func in agg_funcs:
                                lst.append(agg_func)
                                lst_names.append(agg_names[agg_func] + "_" +value)
                            d[value] = lst
                    # Print NaN values for each value column
                    for value in set(values):
                        cnt = df[value].isnull().sum()
                        logging.debug("NA count for col '"+value+"': "+str(cnt)+" of "+str(df.shape[0]))
                    # AGGREGATE !!
                    df2 = groups.agg(d)

                    # Rename the aggregated columns
                    df2.columns = group_by_dims + lst_names
                # else:
                #     # Pivot table
                #     df2 = pd.pivot_table(df,
                #                          values=values,
                #                          index=group_by_dims,
                #                          aggfunc=[agg_funcs[0]], fill_value=np.NaN, margins=False,
                #                          dropna=True)
                #     # Remove the multiindex in columns
                #     df2.columns = [col[-1] for col in df2.columns.values]
                #     # Remove the index
                #     df2.reset_index(inplace=True)
                # The result, all columns (no index), is stored for later use
                ds = self._create_new_dataset(result_name, ds, df2, group_by_dims, out_names)
            except Exception as e:
                traceback.print_exc()
                issues.append((3, "There was a problem: "+str(e)))

        # Store the dataset in State
        datasets[result_name] = ds

        return issues, None

    def estimate_execution_time(self):
        return 0

    def json_serialize(self):
        # Directly return the metadata dictionary
        return self._content

    def json_deserialize(self, json_input):
        # TODO Read and check keys validity
        issues = []
        if isinstance(json_input, dict):
            self._content = json_input
        else:
            self._content = json.loads(json_input)

        return issues



"""
Evaluation of ASTs

Ideas copied/adapted from:
https://gist.github.com/cynici/5865326

"""
import importlib
import math
import re
from typing import Dict, Tuple, Union, List

import lxml
import numpy as np
import pandas as pd

from nexinfosys import case_sensitive
from nexinfosys.command_generators import global_functions, IType, Issue, IssueLocation, parser_field_parsers
from nexinfosys.command_generators.parser_field_parsers import string_to_ast, arith_boolean_expression, key_value_list, \
    simple_ident, expression_with_parameters, number_interval, arith_boolean_expression_with_less_tokens
from nexinfosys.common.helper import create_dictionary, PartialRetrievalDictionary, strcmp, is_float
from nexinfosys.model_services import State
from nexinfosys.models.musiasem_concepts import ExternalDataset, FactorType, Processor, Hierarchy


# #################################################################################################################### #


# -- FUNCTIONS
def get_interface_type(attribute, value, prd: PartialRetrievalDictionary = None):
    """
    Obtain the name of an InterfaceType given the value of an attribute
    (Obtain the registry of objects)

    :param attribute:
    :param value:
    :param prd: A PartialRetrievalDictionary, passed in State "_glb_idx" to the AST evaluator by
    :return:
    """

    if not prd:
        raise Exception(f"No Global-Index parameter passed to InterfaceType function")
    else:
        # Obtain ALL InterfaceTypes, then ONE having attribute "attribute" with value <value>
        its = prd.get(FactorType.partial_key())
        ret = None
        for it in its:
            v = vars(it).get(attribute)
            if not v:
                v = it.attributes.get(attribute)
            if v and (strcmp(v, str(value)) or (is_float(value) and float(v) == float(value))):
                ret = it.name
                break
        if ret:
            return ret
        else:
            raise Exception(f"No InterfaceType found having attribute '{attribute}' with value '{value}'")


def get_processor(attribute, value, prd: PartialRetrievalDictionary = None):
    """
    Obtain the name of a Processor given the value of an attribute
    (Obtain the registry of objects)

    :param attribute:
    :param value:
    :param prd: A PartialRetrievalDictionary, passed in State "_glb_idx" to the AST evaluator by
    :return:
    """

    if not prd:
        raise Exception(f"No Global-Index parameter passed to Processor function")
    else:
        # Obtain ALL Processors, then ONE having attribute "attribute" with value <value>
        procs = prd.get(Processor.partial_key())
        ret = None
        for proc in procs:
            v = vars(proc).get(attribute)
            if not v:
                v = proc.attributes.get(attribute)
            if v and (strcmp(v, str(value)) or (is_float(value) and float(v) == float(value))):
                ret = proc.name
                break
        if ret:
            return ret
        else:
            raise Exception(f"No Processor found having attribute '{attribute}' with value '{value}'")


def get_nis_name(original_name):
    """
    Convert the original_name to a name valid for NIS

    :param original_name:
    :return:
    """

    prefix = original_name[0] if original_name[0].isalpha() else "_"
    remainder = original_name[1:] if original_name[0].isalpha() else original_name

    return prefix + re.sub("[^0-9a-zA-Z_]", "_", remainder)


def call_udif_function(function_name, state: State = None):
    mod_name, func_name = function_name.rsplit('.', 1)
    mod = importlib.import_module(mod_name)
    func = getattr(mod, func_name)
    kwargs = dict(state=state)
    # CALL FUNCTION!!
    try:
        obj = func(**kwargs)
    except Exception as e:
        obj = None

    return obj


lcia_method_names = {}


def lcia_method(indicator: str, method: str = None, horizon: str = None, compartment: str = None,
                subcompartment: str = None,
                state: State = None, lcia_methods: PartialRetrievalDictionary = None):
    """

    :param indicator: Indicator name
    :param method: LCIA method weighting
    :param horizon: Time horizon
    :param compartment: Compartment to which the indicator applies
    :param subcompartment: Subcompartment to which the indicator applies
    :param state: Current values of processor plus parameters
    :param lcia_methods: Where LCIA data is collected
    :return: A dictionary with the
    """
    global lcia_method_names
    if indicator is None or indicator.strip() == "":
        return None

    try:
        string_to_ast(simple_ident, indicator)
    except:
        indicator = get_nis_name(indicator)

    k = dict(d=indicator)
    if method:
        k["m"] = method
    if horizon:
        k["h"] = horizon
    if compartment:
        k["c"] = compartment
    if subcompartment:
        k["s"] = subcompartment

    ms = lcia_methods.get(key=k, key_and_value=True)
    indices = create_dictionary()
    for k, v in ms:
        method = k["m"]
        if method not in lcia_method_names:
            tmp = method
            try:
                string_to_ast(simple_ident, method)
            except:
                method = get_nis_name(method)
            lcia_method_names[tmp] = method
        else:
            method = lcia_method_names[method]

        idx_name = f'{indicator}_{method}'
        if k["h"] != "":
            idx_name += f'_{k["h"]}'
        if k["c"] != "":
            idx_name += f'_{get_nis_name(k["c"])}'
        if k["s"] != "":
            idx_name += f'_{get_nis_name(k["s"])}'
        if idx_name in indices:
            lst = indices[idx_name]
        else:
            lst = []
            indices[idx_name] = lst
        lst.append((k["i"], v[0], float(v[1])))

    ifaces = create_dictionary()
    for t in state.list_namespace_variables():
        if not t[0].startswith("_"):
            p = t[1]  # * ureg(iface_unit)
            ifaces[t[0]] = p

    res = dict()
    for name, lst in indices.items():
        interfaces = []
        weights = []  # From "
        for t in lst:
            if t[0] in ifaces:
                v = ifaces[t[0]]  # TODO .to(t[1])
                interfaces.append(v)
                weights.append(t[2])
        # Calculate the value
        ind = np.sum(np.multiply(interfaces, weights))  # * ureg(indicator_unit)
        res[name] = ind

    return res


def obtain_processors(xquery: str = None, processors_dom=None, processors_map=None):
    if xquery:
        try:
            processors = set()
            r = processors_dom.xpath(xquery if case_sensitive else xquery.lower())
            for e in r:
                fname = e.get("fullname")
                if fname:
                    processors.add(processors_map[fname])
                else:
                    pass  # Interfaces...
            return processors
        except lxml.etree.XPathEvalError:
            # TODO Try CSSSelector syntax
            # TODO Generate Issue
            pass
    else:
        # ALL
        # TODO Probably this should not be used because it will normally imply double accounting
        return set(processors_map.values())


def get_adapted_case_dataframe_filter(df, column, values):
    i_names = df.index.unique(level=column).values
    i_names_case = [_ if case_sensitive else _.lower() for _ in i_names]
    i_names_corr = dict(zip(i_names_case, i_names))
    # https://stackoverflow.com/questions/18453566/python-dictionary-get-list-of-values-for-list-of-keys
    return [i_names_corr[_] for _ in values.intersection(i_names_case)]


def obtain_subset_of_processors(processors_selector: str, serialized_model: lxml.etree._ElementTree,
                                registry: PartialRetrievalDictionary,
                                p_map: Dict[str, Processor], df: Union[List, pd.DataFrame]) -> pd.DataFrame:
    processors = obtain_processors(processors_selector, serialized_model, p_map)

    if len(p_map) == len(processors):
        processors = set()

    if isinstance(df, pd.DataFrame):
        dfs = [df]
    else:
        dfs = df

    results = []

    # Filter Processors
    if len(processors) > 0:
        # Obtain names of processor to KEEP
        processor_names = set([_.full_hierarchy_names(registry)[0] for _ in processors])
        if not case_sensitive:
            processor_names = set([_.lower() for _ in processor_names])

        p_names = get_adapted_case_dataframe_filter(df, "Processor", processor_names)
        # p_names = df.index.unique(level="Processor").values
        # p_names_case = [_ if case_sensitive else _.lower() for _ in p_names]
        # p_names_corr = dict(zip(p_names_case, p_names))
        # # https://stackoverflow.com/questions/18453566/python-dictionary-get-list-of-values-for-list-of-keys
        # p_names = [p_names_corr[_] for _ in processor_names.intersection(p_names_case)]
        # Filter dataframe to only the desired Processors

        for df_ in dfs:
            results.append(df_.query('Processor in [' + ', '.join(['"' + p + '"' for p in p_names]) + ']'))
    else:
        for df_ in dfs:
            results.append(df_)
        processors = p_map

    if isinstance(df, pd.DataFrame):
        results = results[0]

    return results, processors


def aggregator_generic(funct, field: str, xquery: str = None, scope: str = 'Total', processors_dom=None,
                       processors_map=None, df_group=None, df_indicators_group=None):
    """
    AGGREGATE "field" for all processors meeting the XQuery and scope, applying aggregator "funct"
    :param field:
    :param xquery:
    :param processors_dom:
    :param processors_map:
    :param df_group:
    :return:
    """
    inv_map = {v: k for k, v in processors_map.items()}
    processors = obtain_processors(xquery, processors_dom, processors_map)
    processor_names = set([inv_map[p] for p in processors])
    # Only one Interface(and its Orientation) allowed
    # Filter the passed group by processor and scope, by Interface and Orientation
    # Aggregate the Value column according of remaining rows
    f = field.lower()
    orient = ["Input", "Output"]
    if field.lower().endswith("_input"):
        f = f[:-len("_input")]
        orient = ["Input"]
    elif field.lower().endswith("_output"):
        f = f[:-len("_output")]
        orient = ["Output"]
    filt = df_group.index.get_level_values("Processor").isin(processor_names) & \
           df_group.index.get_level_values("Scope").isin([scope]) & \
           df_group.index.get_level_values("Interface").isin([f]) & \
           df_group.index.get_level_values("Orientation").isin(orient)
    values = df_group[filt]["Value"].values
    if len(values) > 0:
        return funct(values)
    else:
        return None


def aggregator_sum(field: str, xquery: str = None, scope: str = 'Total', processors_dom=None, processors_map=None,
                   df_group=None, df_indicators_group=None):
    return aggregator_generic(np.nansum, field, xquery, scope, processors_dom, processors_map, df_group,
                              df_indicators_group)


def aggregator_avg(field: str, xquery: str = None, scope: str = 'Total', processors_dom=None, processors_map=None,
                   df_group=None, df_indicators_group=None):
    return aggregator_generic(np.nanavg, field, xquery, scope, processors_dom, processors_map, df_group,
                              df_indicators_group)


def aggregator_max(field: str, xquery: str = None, scope: str = 'Total', processors_dom=None, processors_map=None,
                   df_group=None, df_indicators_group=None):
    return aggregator_generic(np.nanmax, field, xquery, scope, processors_dom, processors_map, df_group,
                              df_indicators_group)


def aggregator_min(field: str, xquery: str = None, scope: str = 'Total', processors_dom=None, processors_map=None,
                   df_group=None, df_indicators_group=None):
    return aggregator_generic(np.nanmin, field, xquery, scope, processors_dom, processors_map, df_group,
                              df_indicators_group)


def aggregator_count(field: str, xquery: str = None, scope: str = 'Total', processors_dom=None, processors_map=None,
                     df_group=None, df_indicators_group=None):
    return aggregator_generic(lambda v: len(v), field, xquery, scope, processors_dom, processors_map, df_group,
                              df_indicators_group)


def aggregator_nan_count(field: str, xquery: str = None, scope: str = 'Total', processors_dom=None, processors_map=None,
                         df_group=None, df_indicators_group=None):
    return aggregator_generic(lambda v: sum(np.isnan(v)), field, xquery, scope, processors_dom, processors_map,
                              df_group, df_indicators_group)


# Comparison operators
opMap = {
    "<": lambda a, b: a < b,
    "<=": lambda a, b: a <= b,
    ">": lambda a, b: a > b,
    ">=": lambda a, b: a >= b,
    "==": lambda a, b: a == b,
    "=": lambda a, b: a == b,
    "!=": lambda a, b: a != b,
    "<>": lambda a, b: a != b,
}

na_ops = ("+", "-", "u+", "u-")  # Operators admitting evalution at nest level 0


def ast_evaluator(exp: Dict, state: State, obj, issue_lst, atomic_h_names=False,
                  allowed_functions=global_functions, account_nas_name: str = None,
                  nest_level=0) -> Union[Tuple[float, List[str]], Tuple[str, float, List[str]]]:
    """
    Numerically evaluate the result of the parse of "expression" rule (not valid for the other "expression" rules)

    :param exp: Dictionary representing the AST (output of "string_to_ast" function)
    :param state: "State" used to obtain variables/objects
    :param obj: An object used when evaluating hierarchical variables. simple names, functions and datasets are considered members of this object
    :param issue_lst: List in which issues have to be annotated
    :param atomic_h_names: If True, treat variable names as atomic (False processes them part by part, from left to right). Used in dataset expansion
    :param allowed_functions: Set of allowed function names (and their definitions)
    :param account_nas_name: If True, each evaluation at top level returns 4 results: the evaluation, number of variables,
                        number of not available and number of not applicable (i.e., like zero)
    :param nest_level: Recursive level of the function. Used to allow evaluation of addends when they are NotAvailable
    :return: value (scalar EXCEPT for named parameters, which return a tuple "parameter name - parameter value"), list of unresolved variables
    """
    ret_var = None
    ret_val = None
    # TODO dict()
    #      Resolved vars will be "Av"
    #      Each unresolved var will be marked as "NAv" (not-available) or "NAp" (not-applicable, with counts as zero)
    involved_vars = dict()
    if "type" in exp:
        t = exp["type"]
        if t in ("int", "float", "str", "boolean"):  # Literals
            ret_val = exp["value"]
        elif t == "key_value_list":
            d = create_dictionary()
            for k, v in exp["parts"].items():
                d[k], tmp = ast_evaluator(v, state, obj, issue_lst, atomic_h_names, allowed_functions,
                                          account_nas_name=account_nas_name, nest_level=nest_level + 1)
                involved_vars.update(tmp)
            ret_val = d
        elif t == "named_parameter":  # Named parameters, special!
            # This one returns a tuple (parameter name, parameter value, unresolved variables)
            v, tmp = ast_evaluator(exp["value"], state, obj, issue_lst, atomic_h_names,
                                   account_nas_name=account_nas_name, nest_level=nest_level + 1)
            involved_vars.update(tmp)
            # Name of the parameter, its value, then "unresolved_vars"
            ret_var = exp["param"]
            ret_val = v
        elif t == "dataset":
            # Function parameters and Slice parameters
            func_params = [ast_evaluator(p, state, obj, issue_lst, atomic_h_names, allowed_functions,
                                         account_nas_name=account_nas_name, nest_level=nest_level + 1) for p in
                           exp["func_params"]]
            slice_params = [ast_evaluator(p, state, obj, issue_lst, atomic_h_names, allowed_functions,
                                          account_nas_name=account_nas_name, nest_level=nest_level + 1) for p in
                            exp["slice_params"]]

            # Find dataset named "exp["name"]"
            if obj is None:
                # Global dataset
                ds = state.get(exp["name"], exp["ns"])
                if not ds:
                    issue_lst.append((3, "Global dataset '" + exp["name"] + "' not found"))
            else:
                # Dataset inside "obj"
                try:
                    ds = getattr(obj, exp["name"])
                except:
                    ds = None
                if not ds:
                    issue_lst.append((3, "Dataset '" + exp["name"] + "' local to " + str(obj) + " not found"))

            involved_vars = None
            if ds and isinstance(ds, ExternalDataset):
                ret_val = ds.get_data(None, slice_params, None, None, func_params)
            else:
                ret_val = None
        elif t == "function":  # Call function
            # First, obtain the Parameters
            args = []
            kwargs = {}
            can_resolve = True
            for p in [ast_evaluator(p, state, obj, issue_lst, atomic_h_names, allowed_functions,
                                    account_nas_name=account_nas_name, nest_level=nest_level + 1) for p in
                      exp["params"]]:
                if len(p) == 3:
                    kwargs[p[0]] = p[1]
                    tmp = p[2]
                else:
                    args.append(p[0])
                    tmp = p[1]
                involved_vars.update(tmp)
                if len(tmp) > 0:
                    can_resolve = False

            if obj is None:
                # Check if it can be resolved (all variables specified)
                # Check if global function exists, then call it. There are no function namespaces (at least for now)
                if can_resolve and exp["name"] in allowed_functions:
                    _f = allowed_functions[exp["name"]]
                    mod_name, func_name = _f["full_name"].rsplit('.', 1)
                    mod = importlib.import_module(mod_name)
                    func = getattr(mod, func_name)
                    kwargs.update(_f["kwargs"])
                    for sp_kwarg, name in _f.get("special_kwargs", {}).items():
                        if sp_kwarg == "PartialRetrievalDictionary":
                            kwargs[name] = state.get("_glb_idx")
                        elif sp_kwarg == "ProcessorsDOM":
                            kwargs[name] = state.get("_processors_dom")
                        elif sp_kwarg == "ProcessorsMap":
                            kwargs[name] = state.get("_processors_map")
                        elif sp_kwarg == "DataFrameGroup":
                            kwargs[name] = state.get("_df_group")
                        elif sp_kwarg == "IndicatorsDataFrameGroup":
                            kwargs[name] = state.get("_df_indicators_group")
                        elif sp_kwarg == "IndicatorState":
                            kwargs[name] = state
                        elif sp_kwarg == "LCIAMethods":
                            kwargs[name] = state.get("_lcia_methods")
                        elif sp_kwarg == "AccountNA":
                            kwargs[name] = account_nas_name

                    # CALL FUNCTION!!
                    try:
                        obj = func(*args, **kwargs)
                    except Exception as e:
                        obj = None
                        issue_lst.append(str(e))
            else:
                # CALL FUNCTION LOCAL TO THE OBJECT (a "method")
                try:
                    obj = getattr(obj, exp["name"])
                    obj = obj(*args, **kwargs)
                except Exception as e:
                    obj = None
                    issue_lst.append(str(e))
            ret_val = obj
        elif t == "h_var":
            # Evaluate in sequence
            obj = None
            _namespace = exp.get("ns", None)
            if atomic_h_names:
                h_name = '.'.join(exp["parts"])
                exp["parts"] = [h_name]

            for o in exp["parts"]:
                if isinstance(o, str):
                    # Simple name
                    if obj is None:
                        obj = state.get(o, _namespace)
                        if obj is None:
                            issue_lst.append((3, "'" + o + "' is not globally declared in namespace '" + (
                                _namespace if _namespace else "default") + "'"))
                            if _namespace:
                                involved_vars[_namespace + "::" + o] = "NAp"
                            else:
                                involved_vars[o] = "NAp"
                            if account_nas_name:
                                obj = 0  # If not defined, the value is zero
                        else:  # Value defined, check if it is a NAv (NaN)
                            if isinstance(obj, (float, int)):  # AccountNAs is valid only for float's
                                if math.isnan(obj):
                                    involved_vars[o] = "NAv"
                                else:
                                    if account_nas_name:
                                        involved_vars[o] = "Av"
                    else:
                        if isinstance(obj, ExternalDataset):
                            # Check if "o" is column (measure) or dimension
                            if o in obj.get_columns() or o in obj.get_dimensions():
                                obj = obj.get_data(o, None)
                            else:
                                issue_lst.append((3, "'" + o + "' is not a measure or dimension of the dataset."))
                        else:
                            try:
                                obj = getattr(obj, o)
                            except:
                                issue_lst.append((3, "'" + o + "' is not a ."))
                else:
                    # Dictionary: function call or dataset access
                    if obj is None:
                        o["ns"] = _namespace
                    obj = ast_evaluator(o, state, obj, issue_lst, atomic_h_names, allowed_functions,
                                        account_nas_name=account_nas_name, nest_level=nest_level + 1)
            if obj is None or isinstance(obj, (str, int, float, bool)):
                ret_val = obj
            # TODO elif isinstance(obj, ...) depending on core object types, invoke a default method, or
            #  issue ERROR if it is not possible to cast to something simple
            else:
                ret_val = obj
        elif t == "condition":  # Evaluate IF part to a Boolean. If True, return the evaluation of the THEN part; if False, return None
            if_result, tmp = ast_evaluator(exp["if"], state, obj, issue_lst, atomic_h_names, allowed_functions)
            involved_vars.update(tmp)
            if len(tmp) == 0 and if_result:
                then_result, tmp = ast_evaluator(exp["then"], state, obj, issue_lst,
                                                 atomic_h_names, allowed_functions,
                                                 account_nas_name=account_nas_name, nest_level=nest_level + 1)
                involved_vars.update(tmp)
                if len(tmp) > 0:
                    then_result = None
                ret_val = then_result
        elif t == "conditions":
            ret_val = None
            for c in exp["parts"]:
                cond_result, tmp = ast_evaluator(c, state, obj, issue_lst, atomic_h_names,
                                                 allowed_functions, account_nas_name=account_nas_name,
                                                 nest_level=nest_level + 1)
                involved_vars.update(tmp)
                if len(tmp) == 0 and cond_result:
                    ret_val = cond_result
                    break
        elif t == "reference":
            ret_val = "[" + exp["ref_id"] + "]"  # TODO Return a special type
        elif t in ("u+", "u-", "exponentials", "multipliers", "adders", "comparison", "not", "and",
                   "or"):  # Arithmetic and Boolean
            # Evaluate recursively the left and right operands
            locally_involved_vars = {}
            if t in ("u+", "u-"):
                current = 0
                # NOTE: Unary operators do not have "left" side. So leave "locally_involved_vars" empty
            else:
                current, tmp1 = ast_evaluator(exp["terms"][0], state, obj, issue_lst, atomic_h_names,
                                              allowed_functions, account_nas_name=account_nas_name,
                                              nest_level=nest_level + 1)
                locally_involved_vars.update(tmp1)
                involved_vars.update(tmp1)

            for i, e in enumerate(exp["terms"][1:]):
                following, tmp2 = ast_evaluator(e, state, obj, issue_lst, atomic_h_names,
                                                allowed_functions, account_nas_name=account_nas_name,
                                                nest_level=nest_level + 1)
                locally_involved_vars.update(tmp2)
                involved_vars.update(tmp2)

                # Type casting for primitive types
                # TODO For Object types, apply default conversion. If both sides are Object, assume number
                if (isinstance(current, (int, float)) and isinstance(following, (int, float))) or \
                        (isinstance(current, bool) and isinstance(following, bool)) or \
                        (isinstance(current, str) and isinstance(following, str)):
                    pass  # Do nothing
                else:  # In others cases, CAST to the operand of the left. This may result in an Exception
                    if current is not None and following is not None:
                        following = type(current)(following)

                op = exp["ops"][i].lower()
                # Either evaluation of terms was ok OR
                # NAs accounting is enabled, nest-level==0 and is an NA compatible operator
                admissible_evaluation = (len([k for k, v in locally_involved_vars.items() if v != "Av"]) == 0) or \
                                        (nest_level == 0 and op in na_ops and account_nas_name is not None)
                if admissible_evaluation:
                    if op in ("+", "-", "u+", "u-"):
                        if current is None:
                            current = 0
                        if following is None:
                            following = 0
                        if op in ("-", "u-"):
                            following = -following
                        current += following
                    elif op in ("*", "/", "//", "%", "**", "^"):
                        if following is None:
                            following = 1
                        if current is None:
                            current = 1
                        if op == "*":
                            current *= following
                        elif op == "/":
                            current /= following
                        elif op == "//":
                            current //= following
                        elif op == "%":
                            current %= following
                        elif op in ("**", "^"):
                            current ^= following
                    elif op == "not":
                        current = not bool(following)
                    elif op == "and":
                        current = current and following
                    elif op == "or":
                        current = current or following
                    else:  # Comparators
                        fn = opMap[op]
                        current = fn(current, following)
                else:
                    current = None  # Could not evaluate because there are missing variables

            if account_nas_name and nest_level == 0:
                # Calculate NAv, NAp and N
                nav = set()
                nap = set()
                av = set()
                m = {"NAv": nav, "NAp": nap, "Av": av}
                # Split in sets depending on the category
                [m[v].add(k) for k, v in involved_vars.items()]
                # Prepare return value
                current = {account_nas_name: current,
                           f"{account_nas_name}_nav": f"{len(nav)}",
                           f"{account_nas_name}_nap": f"{len(nap)}",
                           f"{account_nas_name}_n": f"{len(nav)+len(nap)+len(av)}"}
            elif len([k for k, v in locally_involved_vars.items() if v != "Av"]) > 0:
                current = None

            ret_val = current
        else:
            issue_lst.append((3, "'type' = " + t + " not supported."))
    else:
        issue_lst.append((3, "'type' not present in " + str(exp)))

    # Convert "involved_vars" returned by top level to a set of "unresolved_vars"
    if involved_vars and nest_level == 0:
        involved_vars = {k for k, v in involved_vars.items() if v != "Av"}

    if ret_var is None:
        if involved_vars is not None:
            return ret_val, involved_vars
        else:
            return ret_val  # Dataset
    else:
        return ret_var, ret_val, involved_vars  # Named parameter


def ast_evaluator_static(exp: Dict, state: State, obj, issue_lst, atomic_h_names=False,
                         allowed_functions=global_functions, account_nas_name: str = None,
                         nest_level=0) -> Union[Tuple[float, List[str]], Tuple[str, float, List[str]]]:
    """
    NOT USED
    Statically evaluate the result of the parse of "expression" rule (not valid for the other "expression" rules)
    Return True if the expression can be evaluated (explicitly mentioned variables are defined previously)

    :param exp: Dictionary representing the AST (output of "string_to_ast" function)
    :param state: "State" used to obtain variables/objects
    :param obj: An object used when evaluating hierarchical variables. simple names, functions and datasets are considered members of this object
    :param issue_lst: List in which issues have to be annotated
    :param atomic_h_names: If True, treat variable names as atomic (False processes them part by part, from left to right). Used in dataset expansion
    :param allowed_functions: Set of allowed function names (and their definitions)
    :param account_nas_name: If True, each evaluation at top level returns 4 results: the evaluation, number of variables,
                        number of not available and number of not applicable (i.e., like zero)
    :param nest_level: Recursive level of the function. Used to allow evaluation of addends when they are NotAvailable
    :return: value (scalar EXCEPT for named parameters, which return a tuple "parameter name - parameter value"), list of unresolved variables
    """
    ret_val = None
    # TODO dict()
    #      Each unresolved var will be marked as "NAv" (not-available) or "NAp" (not-applicable, with counts as zero)
    unresolved_vars = set()
    if "type" in exp:
        t = exp["type"]
        if t == "dataset":
            # Function parameters and Slice parameters
            func_params = [ast_evaluator_static(p, state, obj, issue_lst, atomic_h_names, allowed_functions,
                                                account_nas_name=account_nas_name, nest_level=nest_level + 1) for p in
                           exp["func_params"]]
            slice_params = [ast_evaluator_static(p, state, obj, issue_lst, atomic_h_names, allowed_functions,
                                                 account_nas_name=account_nas_name, nest_level=nest_level + 1) for p in
                            exp["slice_params"]]

            # Find dataset named "exp["name"]"
            if obj is None:
                # Global dataset
                ds = state.get(exp["name"], exp["ns"])
                if not ds:
                    issue_lst.append((3, "Global dataset '" + exp["name"] + "' not found"))
                else:
                    ds = True
            else:
                ds = True  # We cannot be sure it will be found, but do not break the evaluation
            # True if the Dataset is True, and the parameters are True
            return ds and all(func_params) and all(slice_params)
        elif t == "function":  # Call function
            # First, obtain the Parameters
            args = []
            kwargs = {}
            can_resolve = True
            for p in [ast_evaluator_static(p, state, obj, issue_lst, atomic_h_names, allowed_functions,
                                           account_nas_name=account_nas_name, nest_level=nest_level + 1) for p in
                      exp["params"]]:
                if len(p) == 3:
                    kwargs[p[0]] = p[1]
                    tmp = p[2]
                else:
                    args.append(p[0])
                    tmp = p[1]
                unresolved_vars.update(tmp)
                if len(tmp) > 0:
                    can_resolve = False

            if obj is None:
                # Check if global function exists, then call it. There are no function namespaces (at least for now)
                if exp["name"] in allowed_functions:
                    _f = allowed_functions[exp["name"]]
                    mod_name, func_name = _f["full_name"].rsplit('.', 1)
                    mod = importlib.import_module(mod_name)
                    func = getattr(mod, func_name)
                    # True if everything is True: function defined and all parameters are True
                    obj = func and all(args) and all(kwargs.values())
            else:
                # Call local function (a "method")
                obj = True
            return obj
        elif t in ("u+", "u-", "exponentials", "multipliers", "adders", "comparison", "not", "and",
                   "or"):  # Arithmetic and Boolean
            # Evaluate recursively the left and right operands
            if t in ("u+", "u-"):
                current = True
                tmp1 = []  # Unary operators do not have "left" side. So empty list for unresolved vars
            else:
                current, tmp1 = ast_evaluator_static(exp["terms"][0], state, obj, issue_lst, atomic_h_names,
                                                     allowed_functions, account_nas_name=account_nas_name,
                                                     nest_level=nest_level + 1)
                unresolved_vars.update(tmp1)

            for i, e in enumerate(exp["terms"][1:]):
                following, tmp2 = ast_evaluator_static(e, state, obj, issue_lst, atomic_h_names,
                                                       allowed_functions, account_nas_name=account_nas_name,
                                                       nest_level=nest_level + 1)
                unresolved_vars.update(tmp2)
                current = current and following

            if len(unresolved_vars) > 0:
                current = None

            ret_val = current
        else:
            issue_lst.append((3, "'type' = " + t + " not supported."))
    else:
        issue_lst.append((3, "'type' not present in " + str(exp)))

    return unresolved_vars


def ast_to_string(exp):
    """
    Elaborate string from expression AST

    :param exp: Input dictionary
    :return: value (scalar EXCEPT for named parameters, which return a tuple "parameter name - parameter value"
    """
    val = None
    if "type" in exp:
        t = exp["type"]
        if t in ("int", "float", "str"):
            val = str(exp["value"])
        elif t == "named_parameter":
            val = str(exp["param"] + "=" + ast_to_string(exp["value"]))
        elif t == "pf_name":
            val = str(exp["processor"] if exp["processor"] else "") + (":" + exp["factor"]) if exp["factor"] else ""
        elif t == "dataset":
            # Function parameters and Slice parameters
            func_params = [ast_to_string(p) for p in exp["func_params"]]
            slice_params = [ast_to_string(p) for p in exp["slice_params"]]

            val = exp["name"]
            if func_params:
                val += "(" + ", ".join(func_params) + ")"
            if slice_params:
                val += "[" + ", ".join(slice_params) + "]"
        elif t == "function":  # Call function
            # First, obtain the Parameters
            val = exp["name"]
            params = []
            for p in [ast_to_string(p) for p in exp["params"]]:
                if isinstance(p, tuple):
                    params.append(p[0] + "=" + p[1])
                else:
                    params.append(p)
            val += "(" + ", ".join(params) + ")"
        elif t == "h_var":
            # Evaluate in sequence
            _namespace = exp["ns"] if "ns" in exp else None
            if _namespace:
                val = _namespace + "::"
            else:
                val = ""

            parts = []
            for o in exp["parts"]:
                if isinstance(o, str):
                    parts.append(o)
                else:
                    # Dictionary: function call or dataset access
                    parts.append(ast_to_string(o))
            val += ".".join(parts)
        elif t in ("u+", "u-", "multipliers", "adders"):  # Arithmetic OPERATIONS
            # Evaluate recursively the left and right operands
            if t in "u+":
                current = ""
            elif t == "u-":
                current = "-"
            else:
                current = ast_to_string(exp["terms"][0])

            for i, e in enumerate(exp["terms"][1:]):
                following = ast_to_string(e)

                op = exp["ops"][i]
                if op in ("+", "-", "u+", "u-"):
                    if current is None:
                        current = 0
                    if following is None:
                        following = 0
                    if op in ("-", "u-"):
                        following = "-(" + following + ")"

                    current = "(" + current + ") + (" + following + ")"
                elif op in ("*", "/", "//", "%"):  # Multipliers
                    if following is None:
                        following = "1"
                    if current is None:
                        current = "1"
                    if op == "*":
                        current = "(" + current + ") * (" + following + ")"
                    elif op == "/":
                        current = "(" + current + ") / (" + following + ")"
                    elif op == "//":
                        current = "(" + current + ") // (" + following + ")"
                    elif op == "%":
                        current = "(" + current + ") % (" + following + ")"
                elif op == "not":
                    if following is None:
                        following = "True"
                    current = "Not (" + following + ")"
                else:  # And, Or, Comparators
                    if following is None:
                        following = "True"
                    if current is None:
                        current = "True"
                    current = "(" + current + ") " + op + "(" + following + ")"

            val = current

    return val


def dictionary_from_key_value_list(kvl, state: State = None):
    """
    From a string containing a list of keys and values, return a dictionary
    Keys must be literals, values can be expressions, to be evaluated at a later moment

    (syntactic validity of expressions is not checked here)

    :param kvl: String containing the list of keys and values
    :except If syntactic problems occur
    :return: A dictionary
    """
    try:
        ast = parser_field_parsers.string_to_ast(key_value_list, kvl)
    except:
        raise Exception(f"Could not parse key-value list: {key_value_list}")

    d = create_dictionary()
    for k, v in ast["parts"].items():
        issues = []
        res, unres = ast_evaluator(v, state, None, issues)
        if len(unres) == 0:
            v = res
            d[k] = v
        else:
            raise Exception(f"Could not evaluate key '{k}' in key-value list. Value: {v}")

    # pairs = kvl.split(",")
    # for p in pairs:
    #     k, v = p.split("=", maxsplit=1)
    #     if not k:
    #         raise Exception(
    #             "Each key-value pair must be separated by '=' and key has to be defined, value can be empty: " + kvl)
    #     else:
    #         try:
    #             k = k.strip()
    #             v = v.strip()
    #             string_to_ast(simple_ident, k)
    #             try:
    #                 # Simplest: string
    #                 string_to_ast(quotedString, v)
    #                 v = v[1:-1]
    #             except:
    #                 issues = []
    #                 ast = string_to_ast(expression_with_parameters, v)
    #                 res, unres = ast_evaluator(ast, state, None, issues)
    #                 if len(unres) == 0:
    #                     v = res
    #
    #             d[k] = v
    #         except:
    #             raise Exception("Key must be a string: " + k + " in key-value list: " + kvl)
    return d


# Check value domain (according to Parameter definition)
def check_parameter_value(glb_idx, p, value, issues, sheet_name, row):
    retval = True
    if p.range:
        try:  # Try "numeric interval"
            ast = string_to_ast(number_interval, p.range)
            # try Convert value to float
            ast2 = string_to_ast(expression_with_parameters, value)
            evaluation_issues: List[Tuple[int, str]] = []
            s = State()
            value, unresolved_vars = ast_evaluator(exp=ast2, state=s, obj=None, issue_lst=evaluation_issues)
            if value is not None:
                try:
                    value = float(value)
                    left = ast["left"]
                    right = ast["right"]
                    left_number = ast["number_left"]
                    right_number = ast["number_right"]
                    if left == "[":
                        value_meets_left = value >= left_number
                    else:
                        value_meets_left = value > left_number
                    if right == "]":
                        value_meets_right = value <= right_number
                    else:
                        value_meets_right = value < right_number
                    if not value_meets_left or not value_meets_right:
                        issues.append(Issue(itype=IType.ERROR,
                                            description=f"The value {value} specified for the parameter '{p.name}' is out of the range {p.range}",
                                            location=IssueLocation(sheet_name=sheet_name, row=row, column=None)))
                        retval = False
                except:
                    issues.append(Issue(itype=IType.ERROR,
                                        description=f"The parameter '{p.name}' has a non numeric value '{value}', and has been constrained with a numeric range. Please, either change the Value or the Range",
                                        location=IssueLocation(sheet_name=sheet_name, row=row, column=None)))
                    retval = False
            else:
                pass  # The parameter depends on other parameters, a valid situation

        except:  # A hierarchy name
            h = glb_idx.get(Hierarchy.partial_key(p.range))
            h = h[0]
            if value not in h.codes.keys():
                issues.append(Issue(itype=IType.ERROR,
                                    description=f"The value '{value}' specified for the parameter '{p.name}' is not in the codes of the hierarchy '{p.range}': {', '.join(h.codes.keys())}",
                                    location=IssueLocation(sheet_name=sheet_name, row=row, column=None)))
                retval = False

    return retval


if __name__ == '__main__':
    from nexinfosys.model_services import State
    from dotted.collection import DottedDict

    # AST_EVALUATOR APPLICATIONS
    # Dataset expansion: dataset name "." dataset field, operations, functions
    # ProcessorScalings: _get_scale_value
    # check_parameter_value (in this module)
    # back_to_nis_format:get_interfaces, to obtain Interface value
    # global_scalar_indicators
    # local_scalar_indicators
    # evaluate_numeric_expression_with_parameters

    ast = string_to_ast(arith_boolean_expression_with_less_tokens, "200 < capacity_factor and capacity_factor < 400")
    ast2 = string_to_ast(arith_boolean_expression, "200 < capacity_factor and capacity_factor < 400")
    s = State()
    s.set("capacity_factor", 300)
    issues = []
    res2, unres2 = ast_evaluator(ast2, s, None, issues)
    res, unres = ast_evaluator(ast, s, None, issues)

    issues = []
    s = State()
    ex = "level =”N - 1”, farm_type =”GH”, efficiency = 0.3"
    ast = string_to_ast(key_value_list, ex)
    res, unres = ast_evaluator(ast, s, None, issues)
    s.set("Param1", 2.1)
    # s.set("Param", 0.1)
    s.set("Param2", 0.2)
    s.set("p1", 2.3)

    ej = "level='n+1', r=[Ref2019], a=5*p1, c=?p1>3 -> 'T1', p1<=2 -> 'T2', 'T3'?"
    ast = string_to_ast(key_value_list, ej)
    res, unres = ast_evaluator(ast, s, None, issues)

    examples = ["?Param1 > 3 -> 5, Param1 <=3 -> 2?",
                "(Param1 * 3 >= 0.3) AND (Param2 * 2 <= 0.345)",
                "cos(Param*3.1415)",
                "{Param} * 3 >= 0.3",
                "'Hola'",
                "'Hola' + 'Adios'",
                "5 * {Param1}",
                "True",
                "'Hola' + Param1"
                ]
    for e in examples:
        try:
            ast = string_to_ast(arith_boolean_expression, e)
            issues = []
            res, unres = ast_evaluator(ast, s, None, issues)
            print(e + ":: AST: " + str(ast))
            if len(unres) > 0:
                print("Unresolved variables: " + str(unres))
            else:
                print(str(type(res)) + ": " + str(res))
        except Exception as e2:
            print("Incorrect: " + e + ": " + str(e2))

    s.set("HH", DottedDict({"Power": {"p": 34.5, "Price": 2.3}}))
    s.set("EN", DottedDict({"Power": {"Price": 1.5}}))
    s.set("HH", DottedDict({"Power": 25}), "ns2")
    s.set("param1", 0.93)
    s.set("param2", 0.9)
    s.set("param3", 0.96)
    examples = [
        # "EN(p1=1.5, p2=2.3)[d1='C11', d2='C21'].v2",  # Simply sliced Variable Dataset (function call)
        "a_function(p1=2, p2='cc', p3=1.3*param3)",
        "-5+4*2",  # Simple expression #1
        "HH",  # Simple name
        "HH.Power.p",  # Hierarchic name
        "5",  # Integer
        "1.5",  # Float
        "1e-10",  # Float scientific notation
        "(5+4)*2",  # Simple expression #2 (parenthesis)
        "3*2/6",  # Simple expression #3 (consecutive operators of the same kind)
        "'hello'",  # String
        "ns2::HH.Power",  # Hierarchic name from another Namespace
        "HH.Power.Price + EN.Power.Price * param1",
        "EN[d1='C11', d2='C21'].d1",  # Simple Dataset slice
        "b.a_function(p1=2, p2='cc', p3=1.3*param3)",
        "b.EN[d1='C11', d2='C21'].d1",  # Hierachical Dataset slice
        "tns::EN(p1=1.5+param2, p2=2.3 * 0.3)[d1='C11', d2='C21'].v2",  # Simply sliced Variable Dataset (function call)
    ]
    for example in examples:
        print(example)
        res = string_to_ast(arith_boolean_expression, example)
        print(res)
        issues = []
        value, unres = ast_evaluator(res, s, None, issues)
        print(str(type(value)) + ": " + str(value) + "; unresolved: " + unres)



from abc import ABCMeta, abstractmethod
from typing import List, Union
import pandas as pd
import numpy as np

import nexinfosys
from nexinfosys import case_sensitive
from nexinfosys.common.helper import create_dictionary, Memoize2, \
    get_dataframe_copy_with_lowercase_multiindex, strcmp
from nexinfosys.models.statistical_datasets import DataSource, Database, Dataset
from nexinfosys.models.musiasem_methodology_support import force_load


class IDataSourceManager(metaclass=ABCMeta):
    @abstractmethod
    def get_name(self) -> str:
        """ Source name """
        pass

    @abstractmethod
    def get_datasource(self) -> DataSource:
        """ Data source """
        pass

    @abstractmethod
    def get_databases(self) -> List[Database]:
        """ List of databases in the data source """
        pass

    @abstractmethod
    def get_datasets(self, database=None) -> list:
        """ List of datasets in a database, or in all the datasource (if database==None)
            Return a list of tuples (database, dataset)
        """
        pass

    @abstractmethod
    def get_dataset_structure(self, database, dataset) -> Dataset:
        """ Obtain the structure of a dataset: concepts, dimensions, attributes and measures """
        pass

    @abstractmethod
    def etl_full_database(self, database=None, update=False):
        """ If bulk download is supported, refresh full database """
        pass

    @abstractmethod
    def etl_dataset(self, dataset, update=False) -> Dataset:
        """ If bulk download is supported, refresh full dataset """
        pass

    @abstractmethod
    def get_dataset_filtered(self, dataset, dataset_params: list) -> Dataset:
        """ Obtains the dataset with its structure plus the filtered values
            The values can be in a pd.DataFrame or in JSONStat compact format
            After this, new dimensions can be joined, aggregations need to be performed
        """
        pass

    @abstractmethod
    def get_refresh_policy(self):  # Refresh frequency for list of databases, list of datasets, and dataset
        pass


class DataSourceManager:

    def __init__(self, session_factory):
        self.registry = create_dictionary()
        self._session_factory = session_factory

    # ---------------------------------------------------------------------------------
    def register_local_datasets(self, local_datasets):
        from nexinfosys.ie_imports.data_sources.ad_hoc_dataset import AdHocDatasets
        # Register AdHocDatasets
        if local_datasets:
            if "adhoc" not in self.registry:
                adhoc = AdHocDatasets(local_datasets)
                self.register_datasource_manager(adhoc)

    def unregister_local_datasets(self, local_datasets):
        from nexinfosys.ie_imports.data_sources.ad_hoc_dataset import AdHocDatasets
        # Unregister AdHocDatasets
        if local_datasets:
            for i in self.registry.values():
                if isinstance(i, AdHocDatasets):
                    self.unregister_datasource_manager(i)
                    break

    def register_datasource_manager(self, instance: IDataSourceManager):
        self.registry[instance.get_name()] = instance

    def unregister_datasource_manager(self, instance: IDataSourceManager):
        """
        Remove a data source
        This is necessary for AdHocDatasets which has to be created and deleted depending on needs
        (it is an Adapter of a map of the Datasets in the current state)

        :param instance:
        :return:
        """
        n = instance.get_name()
        if n in self.registry:
            del self.registry[instance.get_name()]

    def update_data_source(self, source: IDataSourceManager):
        # TODO Clear database
        # TODO Read structure of ALL datasets from this source
        pass

    def _get_source_manager(self, source):
        if source:
            if isinstance(source, str):
                if source in self.registry:
                    source = self.registry[source]
        return source

    # ---------------------------------------------------------------------------------

    def get_supported_sources(self):
        # e.g.: return ["Eurostat", "FAO", "OECD", "FADN", "COMEXT"]
        return [s for s in self.registry]

    def update_sources(self, local_datasets=None):
        """ Update bulk downloads for sources allowing full database. Not full dataset (on demand) """
        # Register AdHoc datasets
        self.register_local_datasets(local_datasets)

        for s in self.registry:
            # Check last update
            poli = self.registry[s].get_refresh_policy()
            update = False
            if update:
                pass

        # Unregister AdHoc datasets
        self.unregister_local_datasets(local_datasets)

    def get_databases(self, source: Union[IDataSourceManager, str], local_datasets=None):
        # Register AdHoc datasets
        self.register_local_datasets(local_datasets)

        # List of databases in source (a database contains one or more datasets)
        if source:
            source = self._get_source_manager(source)

        lst = []
        if not source:
            for s in self.registry:
                lst.extend([(s,
                             [db for db in self.registry[s].get_databases()]
                             )
                            ]
                           )
        else:
            lst = [(source.get_name(),
                     [db for db in source.get_databases()]
                    )
                   ]

        # Unregister AdHoc datasets
        self.unregister_local_datasets(local_datasets)

        return lst

    # @cachier(stale_after=datetime.timedelta(days=1))
    def get_datasets(self, source: Union[IDataSourceManager, str]=None, database=None, local_datasets=None):
        # Register AdHoc datasets
        self.register_local_datasets(local_datasets)

        if source:
            source = self._get_source_manager(source)

        lst = []
        if source:
            if strcmp(source.get_name(), "AdHoc"):
                lst = [(source.get_name(), source.get_datasets(database))]
            else:
                lst = self.get_external_datasets(source, database)
        else:  # ALL DATASETS
            lst_total = []
            lst = self.get_external_datasets(source, database)  # Because "get_external_datasets" uses "Memoize", DO NOT modify "lst" outside
            lst_total.extend(lst)
            for s in self.registry:
                if strcmp(s, "AdHoc") and local_datasets:
                    lst_total.append((s, [ds for ds in self.registry[s].get_datasets()]))

        # Unregister AdHoc datasets
        self.unregister_local_datasets(local_datasets)

        return lst

    @Memoize2
    def get_external_datasets(self, source: Union[IDataSourceManager, str]=None, database=None):
        """
        Obtain a list of tuples (Source, Dataset name)

        :param source: If specified, the name of the source
        :param database: If specified, the name of a database in the source
        :return: List of tuples (Source name, Dataset name)
        """

        if source:
            source = self._get_source_manager(source)

        if source:
            if database:  # SOURCE+DATABASE DATASETS
                return [(source.get_name(), source.get_datasets(database))]
            else:  # ALL SOURCE DATASETS
                lst = []
                for db in source.get_databases():
                    lst.extend(source.get_datasets(db))
                return [(source.get_name(), lst)]  # List of tuples (dataset code, description, urn)
        else:  # ALL DATASETS
            lst = []
            for s in self.registry:
                if not strcmp(s, "AdHoc"):
                    lst.append((s, [ds for ds in self.registry[s].get_datasets()]))
            return lst  # List of tuples (source, dataset code, description, urn)

    def get_dataset_structure(self, source: Union[IDataSourceManager, str], dataset: str, local_datasets=None) -> Dataset:
        """ Obtain the structure of a dataset, a list of dimensions and measures, without data """
        # Register AdHoc datasets
        self.register_local_datasets(local_datasets)

        if not source:
            source = DataSourceManager.obtain_dataset_source(dataset, local_datasets)
            if not source:
                raise Exception("Could not find a Source containing the Dataset '"+dataset+"'")

        source = self._get_source_manager(source)
        struc = source.get_dataset_structure(None, dataset)

        # Unregister AdHoc datasets
        self.unregister_local_datasets(local_datasets)

        return struc

    def get_dataset_filtered(self, source: Union[IDataSourceManager, str], dataset: str, dataset_params: dict, local_datasets=None) -> Dataset:
        """ Obtain the structure of a dataset, and DATA according to the specified FILTER, dataset_params """
        # Register AdHoc datasets
        self.register_local_datasets(local_datasets)
        source = self._get_source_manager(source)

        fds = source.get_dataset_filtered(dataset, dataset_params)

        # Unregister AdHoc datasets
        self.unregister_local_datasets(local_datasets)

        return fds

    @staticmethod
    def obtain_dataset_source(dset_name, local_datasets=None):
        from nexinfosys.ie_imports.data_sources.ad_hoc_dataset import AdHocDatasets
        # Register AdHocDatasets
        if local_datasets:
            if "AdHoc" not in nexinfosys.data_source_manager.registry:
                adhoc = AdHocDatasets(local_datasets)
                nexinfosys.data_source_manager.register_datasource_manager(adhoc)

        # Obtain the list of ALL datasets, and find the desired one, then find the source of the dataset
        lst = nexinfosys.data_source_manager.get_datasets(None, None, local_datasets)  # ALL Datasets, (source, dataset)
        ds = create_dictionary(data={d[0]: t[0] for t in lst for d in t[1]})  # Dataset to Source (to obtain the source given the dataset name)

        if dset_name in ds:
            source = ds[dset_name]
        else:
            source = None

        # Unregister AdHocDatasets
        if local_datasets:
            nexinfosys.data_source_manager.unregister_datasource_manager(adhoc)

        return source

# --------------------------------------------------------------------------------------------------------------------


def get_dataset_structure(session_factory, source: IDataSourceManager, dataset: str) -> Dataset:
    """ Helper function called by IDataSourceManager implementations """

    src_name = source.get_name()

    # ACCESS TO METADATA DATABASE
    session = session_factory()
    # Check if the source exists. Create it if not
    src = session.query(DataSource).filter(DataSource.name == src_name).first()
    if not src:
        src = source.get_datasource()
        session.add(src)
    # Check if the dataset exists. "ETL" it if not
    ds = session.query(Dataset).\
        filter(Dataset.code == dataset).\
        join(Dataset.database).join(Database.data_source).\
        filter(DataSource.name == src_name).first()
    if not ds:
        # >>>> It may imply a full ETL operation <<<<
        ds = source.etl_dataset(dataset, update=False)
        # Use existing database
        db = session.query(Database).filter(Database.code == ds.database.code).first()
        if db:
            ds.database = db
        else:
            ds.database.data_source = src  # Assign the DataSource to the Database
        session.add(ds)

    session.commit()

    force_load(ds)

    session.close()
    session_factory.remove()
    # END OF ACCESS TO DATABASE

    return ds


def filter_dataset_into_dataframe(in_df, filter_dict, dataset_name, eurostat_postprocessing=False):
    """
    Function allowing filtering a dataframe passed as input,
    using the information from "filter_dict", containing the dimension names and the list
    of codes that should pass the filter. If several dimensions are specified an AND combination
    is done

    "in_df" must have as index a MultiIndex with all the dimensions appearing in the "filter_dict"


    :param in_df: Input dataset, pd.DataFrame
    :param filter_dict: A dictionary with the items to keep, per dimension
    :param dataset_name: Original dataset name (to obtain dimensions and their domains)
    :param eurostat_postprocessing: Eurostat dataframe needs special postprocessing. If True, do it
    :return: Filtered dataframe
    """

    # TODO If a join is requested, do it now. Add a new element to the INDEX
    # TODO The filter params can contain a filter related to the new joins

    start = None
    if "StartPeriod" in filter_dict:
        start = filter_dict["StartPeriod"]
        if isinstance(start, list): start = start[0]
    if "EndPeriod" in filter_dict:
        endd = filter_dict["EndPeriod"]
        if isinstance(endd, list): endd = endd[0]
    else:
        if start:
            endd = start
    if not start:
        columns = in_df.columns  # All columns
    else:
        # Assume year, convert to integer, generate range, then back to string
        start = int(start)
        endd = int(endd)
        columns = [str(a) for a in range(start, endd + 1)]

    # Combinatorial dataset
    dset = nexinfosys.data_source_manager.get_dataset_structure(None, dataset_name)
    pre_combined = dict()
    combined = dict()
    for d in dset.dimensions:
        if not d.is_measure:
            if not d.is_time:
                cl = []
                if d.get_hierarchy():
                    if isinstance(d.get_hierarchy, list):
                        for v in d.get_hierarchy().codes:
                            cl.append(v.name)
                    else:  # Fix: codes can be in a dictionary
                        for v in d.get_hierarchy().codes.values():
                            cl.append(v.name)
                pre_combined[d.code] = cl
            else:
                pre_combined[d.code] = columns
    for i, k in enumerate(in_df.index.names):
        if k in filter_dict:
            combined[k] = filter_dict[k]
        else:
            combined[k] = pre_combined[k]

    # To bring all possible cells, prepare a combinatorial then merge with explicit values from dataset
    midx= pd.MultiIndex.from_product(combined.values(), names=combined.keys())
    in_df = in_df.merge(pd.DataFrame(index=midx), left_index=True, right_index=True, how='outer').fillna(np.NaN)

    if not case_sensitive:
        in_df_lower = get_dataframe_copy_with_lowercase_multiindex(in_df)

    # Rows (dimensions)
    cond_accum = np.full(in_df.index.size, fill_value=True)
    for i, k in enumerate(in_df.index.names):
        if k in filter_dict:
            lst = filter_dict[k]
            if not isinstance(lst, list):
                lst = [lst]
            if len(lst) > 0:
                if not case_sensitive:
                    cond_accum &= in_df_lower.index.isin([str(l).lower() for l in lst], i)
                else:
                    cond_accum &= in_df.index.isin([str(l) for l in lst], i)
            else:
                cond_accum &= in_df[in_df.columns[0]] == in_df[in_df.columns[0]]

    # Remove non existent index values
    for v in columns.copy():
        if v not in in_df.columns:
            columns.remove(v)

    tmp = in_df[columns][cond_accum]

    # Convert columns to a single column "TIME_PERIOD"
    if eurostat_postprocessing:
        if len(tmp.columns) > 0:
            lst = []
            for i, cn in enumerate(tmp.columns):
                df2 = tmp[[cn]].copy(deep=True)
                # TODO: use column name from metadata instead of hardcoded "value"
                df2.columns = ["value"]
                df2["TIME_PERIOD"] = cn
                lst.append(df2)
            in_df = pd.concat(lst)
            in_df.reset_index(inplace=True)
            # Value column should be last column
            lst = [l for l in in_df.columns]
            for i, l in enumerate(lst):
                if l == "value":
                    lst[-1], lst[i] = lst[i], lst[-1]
                    break
            in_df = in_df.reindex(lst, axis=1)
            return in_df
        else:
            return None
    else:
        tmp.reset_index(inplace=True)
        if len(tmp.columns) > 0:
            return tmp
        else:
            return None




# ######################################################################################################################
# GRAPH PARTITIONING - SCENARIO (PARAMETERS), SINGLE SCENARIO PARTITION CATEGORIES, IN-SCENARIO OBSERVATIONS VARIATION
# ######################################################################################################################
from model_services import get_case_study_registry_objects
import networkx as nx


def get_contexts(objects):
    """

    :param objects:
    :return:
    """
    pass


def get_graph_partitioning_categories(objects):
    """
    Obtain categories known to partition the graph (into components)

    The most important one is usually "GEO"

    :param objects:
    :return: A list of partition categories
    """
    return ["GEO"]  # TODO Ensure that processors get category "GEO"


# ######################################################################################################################
# PARAMETERS - SCENARIOS. Global parameters. Scenarios to make
# ######################################################################################################################

def get_parameters(objects):
    """
    Obtain all parameters
    Obtain the variation ranges (or constant values) for each parameter

    :param objects:
    :return: A list (or dict) of parameters
    """
    # TODO
    return None


def map_parameters(params, objects):
    """
    Scan all occurrences of parameters
    Put a list in each parameter pointing to the parameter occurrence (expressions)

    :param params:
    :param objects:
    :return:
    """


def get_observation_variation_categories(objects):
    """
    Obtain categories by which, for a given scenario and partition, ie, a decoupled subsystem
    produce variation in quantitative observations, not in the system structure.

    The most important ones are TIME and OBSERVER

    :param objects:
    :return:
    """
    return ["TIME", "SOURCE"]

# ######################################################################################################################
# GENERATORS
# ######################################################################################################################


def get_scenario_generator(params, objects):
    """
    Obtain a scenario generator
    For each iteration it will return a dict with parameter name as key and parameter value as value

    :param params:
    :param objects:
    :return:
    """


def get_partition_generator(scenario, partition_categories, objects):
    """
    Obtain an enumerator of partitions for a given scenario (the set may change from scenario to scenario)
    For each iteration it will return a list of dicts made category names and category values
    (the list may be just of an element)

    :param scenario:
    :param partition_categories:
    :param objects:
    :return:
    """


def get_obs_variation_generator(obs_variation_categories, msm):
    """
    Obtain an iterator on the possible combinations of observation variations, given a MuSIASEM model

    Ideally this would be between one and less than ten [1, 10)
    For each iteration it will return a dict of categories

    :param obs_variation_categories:
    :param msm:
    :return:
    """


# ######################################################################################################################
# DATA STRUCTURES
# ######################################################################################################################


def build_msm_from_parsed(scenario, partition, parse_execution_results):
    """
    Construct MuSIASEM entities given the scenario and the requested partition

    :param scenario:
    :param partition:
    :param parse_execution_results:
    :return:
    """
    # TODO Everything!!!
    # TODO Consider evaluation of parameters in expressions

    return None


def cleanup_unused_processors(msm):
    """
    Remove or disable processors not involved in the solving and indicators
    For instance, unit processors and its children

    Elaborating a mark signaling when we have a unit processor is needed, then filtering out is trivial

    :param msm:
    :return:
    """
    # TODO


def reset_msm_solution_observations(msm):
    """
    Because the solving process in a MuSIASEM model is iterative, and each iteration may add observations,
    a reset is needed before iterations begin.



    :param msm:
    :return:
    """


def put_solution_into_msm(msm, fg):
    """

    :param msm:
    :param fg:
    :return:
    """


def get_flow_graph(msm, obs_variation):
    """
    Obtain a flow graph from the MuSIASEM model
    Assign quantitative observations filtering using the observations variation parameter

    :param msm:
    :param obs_variation:
    :return:
    """
    # TODO
    return nx.DiGraph()


# ######################################################################################################################
# INDICATORS
# ######################################################################################################################


def get_empty_indicators_collector():
    """
    Prepare a data structure where indicators

    It may be an in memory structure or a handle to a persistent structure

    :return:
    """
    # TODO
    return dict()


def compute_local_indicators(msm):
    """
    Compute indicator local to processors

    The standard would be metabolic ratios, wherever it may be possible

    :param msm: Input and output MuSIASEM model, augmented with the local indicators, that would be attached to the Processors
    :return: <nothing>
    """


def compute_global_indicators(msm):
    """
    Compute global indicators -those using data from more than one Processor-
    Store results inside the MuSIASEM structure

    :param msm:
    :return: <Nothing>
    """


def collect_indicators(indicators, msm):
    """
    Extract categories, observations and indicators into the "indicators" analysis structure

    :param indicators:
    :param msm:
    :return: <Nothing>
    """


# ######################################################################################################################
# SOLVER
# ######################################################################################################################


def solve(g):
    """
    Given a flow graph, find the values for the unknown interfaces

    How to treat the situation when an interface, with given value, could get a solution?
    * We have a top-down value
    * A bottom-up value is also available
    * Take note of the solution?

    :param g:
    :return:
    """
    #
    # Elaborate square matrix?




def solver_one(state):
    """
    Solves a MuSIASEM case study AND computes/collects indicators

    Receives as input a registry of parsed/elaborated MuSIASEM objects

    STORES in "state" an indicators structure (a Dataset?) with all the results

    :param state:
    :return: A list of issues
    """

    # Obtain the different state elements
    glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)

    # Obtain parameters and their variation ranges (initially decoupled)
    params_list = get_parameters(state)

    # Map params to the objects where they appear
    map_parameters(params_list, state)  # Modifies "params_list"

    # Obtain "partition" categories
    # Partition categories are categories provoking totally decoupled systems. The most important one is GEO
    partition_categories = get_graph_partitioning_categories(state)

    # Obtain "observation variation" categories
    # Observation variation categories are those which for a given scenario and partition (i.e., fixed subsystem)
    # produce variation in quantitative observations, not in the system structure. The most important are TIME and OBSERVER
    obs_variation_categories = get_observation_variation_categories(state)

    # Empty indicators collector
    indicators = get_empty_indicators_collector()

    for scenario in get_scenario_generator(params_list, state):
        # "scenario" contains a list of parameters and their values
        for partition in get_partition_generator(scenario, partition_categories, state):
            # "partition" contains a list of categories and their specific values
            # Build MSM for the partition categories
            msm = build_msm_from_parsed(scenario, partition, state)
            # TODO Remove processors not in the calculations (unit processors)
            cleanup_unused_processors(msm)  # Modify "msm"
            for obs_variation in get_obs_variation_generator(obs_variation_categories, state):
                # TODO "obs_variation" contains a list of categories and their specific values
                # TODO Build flow graph with observations filtered according to "obs_variation".
                # Nodes keep link to interface AND a value if there is one.
                # Edges keep link to: hierarchy OR flow OR scale change (and context)
                reset_msm_solution_observations(msm)
                fg = get_flow_graph(msm, obs_variation)
                for sub_fg in nx.weakly_connected_component_subgraphs(fg):
                    # Solve the sub_fg. Attach solutions to Nodes of "sub_fg"
                    solve(sub_fg)  # Modify "fg"
                    put_solution_into_msm(msm, sub_fg)  # Modify "msm"
                compute_local_indicators(msm)
                compute_global_indicators(msm)
                collect_indicators(indicators, msm)  # Elaborate output matrices
    return indicators



from io import StringIO

from nexinfosys.model_services import State, get_case_study_registry_objects
from nbformat import write, v4
import nbformat


def generate_jupyter_notebook_python(state: State):
    glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)

    fname = "<get_file_name>"
    nb = v4.new_notebook()
    nb['metadata'] = {
        "kernelspec": {
            "display_name": "Python 3",
            "language": "python",
            "name": "python3"
        },
        "language_info": {
            "codemirror_mode": {
                "name": "ipython",
                "version": 3
            },
            "file_extension": ".py",
            "mimetype": "text/x-python",
            "name": "python",
            "nbconvert_exporter": "python",
            "pygments_lexer": "ipython3",
            "version": "3.7.1"
        }
    }

    nb['cells'] = [
        v4.new_markdown_cell("""
# Jupyter Notebook generated automatically by NIS-frontend
        """),
        v4.new_code_cell("""
from nexinfosys import NISClient, display_visjs_jupyterlab
import io
import pandas as pd
import networkx as nx
        """, ),
        v4.new_code_cell(f"""
fname = "{fname}.xlsx"
            """),
        v4.new_code_cell("""
c = NISClient("https://one.nis.magic-nexus.eu/nis_api")
#c = NISClient("http://localhost:5000/nis_api")

# Login, open session, load a workbook (which is in Nextcloud), submit (execute!)
c.login("test_user")
c.open_session()
#print("Session opened")
n = c.load_workbook(fname)
#print("N worksheets: "+str(n))
r = c.submit()
#print("Returned from submit")
# Check if submission was successful (it should be with the provided workbook), then query 
# available datasets, and get one of them, converting it into a pd.DataFrame
any_error = False
if len(r) > 0:
    for i in r:
        if i["type"] == 3:
            any_error = True
            print(str(i))

if not any_error:
    # Obtain available datasets
    r = c.query_available_datasets()
    if len(r) > 0:
        results = {}
        for ds in r:
            results[ds["name"]] = {d["format"].lower(): d["url"] for d in ds["formats"]}
            #print(str(ds))
        #r = c.download_results([(results["FG"]["visjs"])])
        #visjs_data = r[0].decode("utf-8")
        #unique_name = None
        r = c.download_results([(results["PG"]["visjs"])])
        visjs_data2 = r[0].decode("utf-8")
        un2 = None

        #unique_name = display_visjs_jupyterlab(visjs_data, 800, unique_name)
        un2 = display_visjs_jupyterlab(visjs_data2, 1200, un2)

        """)
    ]

    return nbformat.writes(nb)


def generate_jupyter_notebook_r(state: State):
    glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)

    nb = v4.new_notebook()
    nb['metadata'] = {
        "kernelspec": {
            "display_name": "R",
            "language": "R",
            "name": "ir"
        },
        "language_info": {
            "codemirror_mode": "r",
            "file_extension": ".r",
            "mimetype": "text/x-r-source",
            "name": "R",
            "pygments_lexer": "r",
            "version": "3.5.1"
        }
    }

    nb['cells'] = [
        v4.new_markdown_cell("""
# Jupyter Notebook generated automatically by NIS-frontend
        """),
        v4.new_code_cell("""
.libPaths( c("~/R-packages", "/opt/conda/lib/R/library") )
#install.packages("reticulate")
library("reticulate")
nexinfosys <- import("nexinfosys")
c <- nexinfosys$NISClient("https://one.nis.magic-nexus.eu/nis_api")        
        """),
        v4.new_code_cell("""
fname <- "https://nextcloud.data.magic-nexus.eu/remote.php/webdav/NIS_beta/CS_format_examples/08_caso_energia_eu_new_commands.xlsx"
        """),
        v4.new_code_cell("""
c$login("test_user")
print("Logged in")
c$open_session()
print("Session opened")
n <- c$load_workbook(fname, "NIS_agent", "NIS_agent@1")
print(paste("N worksheets: ",n))
r <- c$submit()
print("Returned from submit")
        """),
        v4.new_code_cell("""
# Obtain a list of available datasets
r <- c$query_available_datasets()
        """),
        v4.new_code_cell("""
# Obtain a specific dataset
ds <- c$query_datasets(c(tuple("ds1", "csv", "dataframe")))
        """),
        v4.new_code_cell("""
# Obtain an R data.frame
df <- py_to_r(ds[[1]][[3]])
        """),
        v4.new_code_cell("""
c$close_session()
c$logout()
        """)
    ]

    return nbformat.writes(nb)